'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

"""This module contains commands for building Arches."""

import os
import sys
from optparse import make_option
from django.core.management.base import BaseCommand, CommandError
from django.db import connection, transaction
import arches.app.models.models as archesmodels
from arches.app.models.concept import Concept
from arches.app.utils.betterJSONSerializer import JSONSerializer, JSONDeserializer
from arches.app.search.search_engine_factory import SearchEngineFactory

class Command(BaseCommand):
    """A general command used index Arches data into Elasticsearch."""

    option_list = BaseCommand.option_list + (
        make_option('-d', '--delete', action='store', dest='delete', default='',
            type='choice', choices=['all', 'maplayers', 'entity', 'term', 'concept', ''],
            help='Delete Type; all=Deletes the "maplayers, entity, term, and concept" indexes from Elasticsearch, ' +
            '\'maplayers\'=Deletes just the maplayers index, ' + 
            '\'entity\'=Deletes just the entity index, ' +
            '\'term\'=Deletes just the term index, '),

        make_option('-i', '--index', action='store', dest='index', default='',
            type='choice', choices=['concept', ''],
            help='Index Type; concept=Indexes just the concepts, '),
    )
    
    def handle(self, *args, **options):
        if options['delete'] != '':
            print 'delete: '+ options['delete']
            if options['delete'] == 'all':
                self.delete_index('maplayers')
                self.delete_index('entity')
                self.delete_index('term')
                self.delete_index('concept')
            else:
                self.delete_index(options['delete'])

        if options['index'] != '':    
            print 'index: '+ options['index']
            cursor = connection.cursor()

            if options['index'] == 'concept' or options['index'] == 'all':
                self.index_concepts_for_search()
                sql = """
                    SELECT a.entitytypeid
                    FROM data.entity_types a;
                    """
                cursor.execute(sql)
                entitytypeids = cursor.fetchall()
                for entitytypeid in entitytypeids:
                    self.index_concepts_by_entitytypeid(entitytypeid[0])

    def index_concepts_for_search(self):
        # see http://sqlblog.com/blogs/adam_machanic/archive/2006/07/12/swinging-from-tree-to-tree-using-ctes-part-1-adjacency-to-nested-sets.aspx          
        # Value of Lft for the root node is 1
        # Value of Rgt for the root node is 2 * (Number of nodes)
        # Value of Lft for any node is ((Number of nodes visited) * 2) - (Level of current node)
        # Value of Rgt for any node is (Lft value) + ((Number of subnodes) * 2) + 1 
     

        sys.setrecursionlimit(3000)
        se = SearchEngineFactory().create()
        se.create_mapping('concept', 'all', 'conceptid', 'string', 'not_analyzed')
        se.create_mapping('concept', 'all', 'labelid', 'string', 'not_analyzed')

        def _findNarrowerConcept(conceptid, ret=None, limit=200000, level=1):
            returnobj = {'subnodes': 0}
            if ret == None: # the root node
                labels = archesmodels.Values.objects.filter(conceptid = conceptid)
                ret = {}
                nodesvisited = len(ret) + 1
                ret[conceptid] = {'labels': [], 'left': (nodesvisited*2)-level, 'right': 0}               
                for label in labels:
                    ret[conceptid]['labels'].append({'labelid': label.pk, 'label': label.value})
                level = level + 1

            conceptrealations = archesmodels.ConceptRelations.objects.filter(conceptidfrom = conceptid)
            for relation in conceptrealations:
                nodesvisited = len(ret) + 1
                labels = archesmodels.Values.objects.filter(conceptid = relation.conceptidto)
                ret[relation.conceptidto_id] = {'labels': [], 'left': (nodesvisited*2)-level, 'right': 0}
                for label in labels:
                    ret[relation.conceptidto_id]['labels'].append({'labelid': label.pk, 'label': label.value})
                returnobj = _findNarrowerConcept(relation.conceptidto_id, ret=ret, level=level+1)      
            
            subnodes = returnobj['subnodes']
            if subnodes == 0: # meaning we're at a leaf node
                ret[conceptid]['right'] = ret[conceptid]['left'] + 1
            else:
                ret[conceptid]['right'] = subnodes + 1
            return {'all_concepts': ret, 'subnodes': ret[conceptid]['right']}

        concepts = _findNarrowerConcept('00000000-0000-0000-0000-000000000003')
        
        all_concepts = []
        for key, concept in concepts['all_concepts'].iteritems():
            all_concepts.append({'conceptid': key, 'labels': concept['labels'], 'left': concept['left'], 'right': concept['right']})

        self.index(all_concepts, 'concept', 'all', 'conceptid')

    def index(self, documents, index, type, idfield, processdoc=None, getid=None, bulk=False):
        detail = ''
        bulkitems = []
        errorlist = []
        se = SearchEngineFactory().create()
        if not isinstance(documents, list):
            documents = [documents]
        for document in documents:
            #print "inserting document: %s" % (document)
            sys.stdout.write('.')
            if processdoc == None:
                data = document
            else:
                data = processdoc(document)
            id = None
            if getid != None:
                id = getid(document, data)            
            try:
                if bulk:
                    bulkitem = se.create_bulk_item(index, type, id, data)
                    bulkitems.append(bulkitem[0])
                    bulkitems.append(bulkitem[1])        
                else:
                    se.index_data(index, type, data, idfield=idfield, id=id)
            except Exception as detail:
                errorlist.append(id)
        if bulk:
            try:
                se.bulk_index(index, type, bulkitems)
            except Exception as detail:
                errorlist = bulkitems
                print 'bulk inset failed'

        if detail != '':
            print "\n\nException detail: %s " % (detail)
            print "There was a problem indexing the following items:"
            print errorlist

    def index_concepts_by_entitytypeid(self, entitytypeid):
        entitytype = archesmodels.EntityTypes.objects.get(pk = entitytypeid)
        conceptid = entitytype.conceptid_id
        concept_graph = Concept().get(id=conceptid, include_subconcepts=True, exclude=['note'])
        if len(concept_graph.subconcepts) > 0:
            data = JSONSerializer().serializeToPython(concept_graph, ensure_ascii=True, indent=4)
            self.index(data, 'concept', entitytypeid, 'id', processdoc=None, getid=None, bulk=False)  

    def delete_index(self, index):
        se = SearchEngineFactory().create()
        se.delete_index(index=index)




'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

from arches.app.utils.betterJSONSerializer import JSONSerializer

class Dsl(object):
    def __init__(self, dsl=None):
        if dsl is None: 
            self.dsl = {}
        else:
            self.dsl = dsl

    def __str__(self):
        return JSONSerializer().serialize(self.dsl, indent=4) 

    @property
    def dsl(self):
        return self._dsl

    @dsl.setter
    def dsl(self, value):
        try:
            self._dsl = value.dsl
        except AttributeError:
            self._dsl = value  

class Query(Dsl):
    """
    http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/query-dsl.html

    """

    def __init__(self, se, **kwargs):
        self.se = se
        self._filtered = False
        self.filter_operator = None
        self.fields = kwargs.pop('fields', None)
        self.start = kwargs.pop('start', 0)
        self.limit = kwargs.pop('limit', 10)

        self.dsl = {
            'query': {
                "match_all": { }
            }
        }

    def add_filter(self, dsl=None, operator='and'):
        """
        Wrap the filter in an "and" or "or" clause by specifying an operator value of "and" or "or"

        """

        if dsl is not None:
            dsl = Dsl(dsl).dsl
            if operator:
                self.filter_operator = operator

            if self._filtered:
                self.dsl['query']['filtered']['filter'].append(dsl)
            else:
                self._filtered = True
                self.dsl = {
                    'query':{
                        'filtered':{
                            'query': self.dsl['query'],
                            'filter': [dsl]
                        }
                    }
                }

    def add_query(self, dsl=None):
        if dsl is not None:
            dsl = Dsl(dsl).dsl

            if self._filtered:
                if 'bool' in dsl and 'bool' in self.dsl['query']['filtered']['query']:
                    self.dsl['query']['filtered']['query'] = Bool(self.dsl['query']['filtered']['query']).merge(dsl).dsl
            else:
                if 'bool' in dsl and 'bool' in self.dsl['query']:
                    self.dsl['query'] = Bool(self.dsl['query']).merge(dsl).dsl
                else:
                    self.dsl['query'] = dsl

    def search(self, index='', doc_type='', **kwargs):
        self.fields = kwargs.pop('fields', self.fields)
        self.start = kwargs.pop('start', self.start)
        self.limit = kwargs.pop('limit', self.limit)

        self.prepare()
        print self
        return self.se.search(index=index, doc_type=doc_type, body=self.dsl)

    def delete(self, index=''):
        return self.se.delete(index=index, body=self.dsl)

    def prepare(self):
        self.dsl['from'] = self.start
        self.dsl['size'] = self.limit

        if self._filtered:
            self.dsl['query']['filtered']['filter'] = {   
                self.filter_operator: self.dsl['query']['filtered']['filter']
            }

        if self.fields != None:
            self.dsl['fields'] = self.fields


class Bool(Dsl):
    """
    http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/query-dsl-bool-query.html
    
    """
    
    def __init__(self, dsl=None, **kwargs):  
        self.dsl = {
            'bool':{
                'should':[],
                'must':[],
                'must_not':[]
            }
        }
        if isinstance(dsl, dict):
            self.dsl['bool'] = dsl['bool']
        elif isinstance(dsl, Bool):
            self = dsl

        self.should(kwargs.pop('should', None))
        self.must(kwargs.pop('must', None))
        self.must_not(kwargs.pop('must_not', None))
        self.empty = True

    def must(self, dsl):
        return self._append('must', dsl)

    def should(self, dsl):
        return self._append('should', dsl)

    def must_not(self, dsl):
        return self._append('must_not', dsl)

    def _append(self, type, dsl):
        if dsl:
            dsl = Dsl(dsl)
            self.dsl['bool'][type].append(dsl.dsl)
            self.empty = False
        return self

    def merge(self, object):
        if isinstance(object, dict):
            object = Bool(object)

        self.dsl['bool']['must'] = self.dsl['bool']['must'] + object.dsl['bool']['must']
        self.dsl['bool']['should'] = self.dsl['bool']['should'] + object.dsl['bool']['should']
        self.dsl['bool']['must_not'] = self.dsl['bool']['must_not'] + object.dsl['bool']['must_not']

        return self
       

class Match(Dsl):
    """
    http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/query-dsl-match-query.html
    
    """

    def __init__(self, **kwargs):
        self.type = kwargs.pop('type', 'boolean')
        self.query = kwargs.pop('query', '')
        self.field = kwargs.pop('field', '_all')
        self.fuzziness = kwargs.pop('fuzziness', None)
        self.operator = kwargs.pop('operator', 'or')
        self.max_expansions = kwargs.pop('max_expansions', 10)

        self.dsl = {
            'match' : {
                self.field : {
                    'query' : self.query,
                    'type' : self.type,
                    'operator': self.operator,
                    'max_expansions' : self.max_expansions
                }
            }
        }

        if self.fuzziness and self.type != 'phrase_prefix':
            self.dsl['match'][self.field]['fuzziness'] = self.fuzziness   

class Nested(Dsl):
    """
    http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/query-dsl-nested-query.html

    Note:
        score_mode can only be used when the nested dsl is used within a query but not within a filter

    """

    def __init__(self, **kwargs):
        self.path = kwargs.pop('path', '')
        self.score_mode = kwargs.pop('score_mode', None)
        self.query = kwargs.pop('query', None)

        self.dsl = {
            'nested' : {
                'path' : self.path
            }
        }

        if self.score_mode:
            self.dsl['nested']['score_mode'] = self.score_mode 

        if self.query:
            self.add_query(dsl=self.query)

    def add_query(self, dsl=None):
        if dsl:
            dsl = Dsl(dsl).dsl
            self.dsl['nested']['query'] = dsl

class Terms(Dsl):
    """
    http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/query-dsl-terms-query.html

    """

    def __init__(self, **kwargs):
        self.field = kwargs.pop('field', '_all')
        self.terms = kwargs.pop('terms', [])

        if not isinstance(self.terms, list):
            self.terms = [self.terms]

        self.dsl = {
            'terms' : {
                self.field : self.terms
            }
        }


class GeoShape(Dsl):
    """
    http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/query-dsl-geo-shape-query.html
    http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/mapping-geo-shape-type.html

    """

    def __init__(self, **kwargs):
        self.field = kwargs.pop('field', '_all')
        self.type = kwargs.pop('type', '')
        self.coordinates = kwargs.pop('coordinates', '')

        self.dsl = {
            'geo_shape': {
                self.field: {
                    'shape': {
                        'type': self.type,
                        'coordinates' : self.coordinates
                    }
                }
            }
        }


class Range(Dsl):
    """
    http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/query-dsl-range-query.html

    """

    def __init__(self, **kwargs):
        self.field = kwargs.pop('field', '_all')
        self.gte = kwargs.pop('gte', None)
        self.gt = kwargs.pop('gt', None)
        self.lte = kwargs.pop('lte', None)
        self.lt = kwargs.pop('lt', None)
        self.boost = kwargs.pop('boost', None)

        if self.boost:
            boost = {
                'boost': self.boost
            }
        else:
            boost = {}
            
        self.dsl = {
            'range' : {
                self.field : boost
            }
        }

        if self.gte is None and self.gt is None and self.lte is None and self.lt is None:
            raise Exception('You need at least one of the following: gte, gt, lte, or lt')
        if self.gte is not None and self.gt is not None:
            raise Exception('You can only use one of either: gte or gt') 
        if self.lte is not None and self.lt is not None:
            raise Exception('You can only use one of either: lte or lt') 
        
        if self.gte is not None:
            self.dsl['range'][self.field]['gte'] = self.gte
        if self.gt is not None:
            self.dsl['range'][self.field]['gt'] = self.gt
        if self.lte is not None:
            self.dsl['range'][self.field]['lte'] = self.lte
        if self.lt is not None:
            self.dsl['range'][self.field]['lt'] = self.lt

class SimpleQueryString(Dsl):
    """
    http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/query-dsl-simple-query-string-query.html

    """

    def __init__(self, **kwargs):
        self.field = kwargs.pop('field', '_all') # can be a list of fields
        self.query = kwargs.pop('query', '')
        self.operator = kwargs.pop('operator', 'or')
        self.analyzer = kwargs.pop('analyzer', 'snowball')
        self.flags = kwargs.pop('flags', 'OR|AND|PREFIX')
        #The available flags are: ALL, NONE, AND, OR, PREFIX, PHRASE, PRECEDENCE, ESCAPE, WHITESPACE, FUZZY, NEAR, and SLOP.

        # if not isinstance(self.field, list):
        #     self.field = [self.field]

        self.dsl = {
            'simple_query_string' : {
                'fields' : self.field,
                'query': self.query,
                'default_operator': self.operator,
                'flags': self.flags
            }
        }

        self.dsl = {
            "prefix" : { self.field : self.query }
        }


'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

# This is an auto-generated Django model module.
# You'll have to do the following manually to clean this up:
#     * Rearrange models' order
#     * Make sure each model has one field with primary_key=True
# Feel free to rename the models, but don't rename db_table values or field names.
#
# Also note: You'll have to insert the output of 'django-admin.py sqlcustom [appname]'
# into your database.

from operator import methodcaller
from django.dispatch import receiver
from django.conf import settings
from django.contrib.gis.db import models
import uuid
import types
import os
import json
from arches.app.utils.betterJSONSerializer import JSONSerializer
from django.contrib.auth.models import User
from django.db.models import Q

# from django.contrib.auth.models import User

# def _func(self):
#     return 'you did it'


# User.get_mapping = types.MethodType(_func, None, User)

class AuthGroup(models.Model):
    id = models.IntegerField(primary_key=True)
    name = models.CharField(unique=True, max_length=80)
    class Meta:
        db_table = u'auth_group'

class AuthUser(models.Model):
    id = models.IntegerField(primary_key=True)
    username = models.CharField(unique=True, max_length=30)
    first_name = models.CharField(max_length=30)
    last_name = models.CharField(max_length=30)
    email = models.CharField(max_length=75)
    password = models.CharField(max_length=128)
    is_staff = models.BooleanField()
    is_active = models.BooleanField()
    is_superuser = models.BooleanField()
    last_login = models.DateTimeField()
    date_joined = models.DateTimeField()
    class Meta:
        db_table = u'auth_user'

class Classes(models.Model):
    classid = models.TextField(primary_key=True)
    classname = models.TextField()
    isactive = models.BooleanField()
    defaultbusinesstable = models.TextField()
    class Meta:
        db_table = u'classes'

    def __unicode__(self):
        return self.classname

class DNodetypes(models.Model):
    nodetype = models.TextField(primary_key=True)
    namespace = models.TextField()
    class Meta:
        db_table = u'd_nodetypes'

class DLanguages(models.Model):
    languageid = models.TextField(primary_key=True)
    languagename = models.TextField()
    isdefault = models.BooleanField()
    class Meta:
        db_table = u'd_languages'

class DRelationtypes(models.Model):
    relationtype = models.TextField(primary_key=True)
    category = models.TextField()
    namespace = models.TextField()
    class Meta:
        db_table = u'd_relationtypes'

class DjangoAdminLog(models.Model):
    id = models.IntegerField(primary_key=True)
    action_time = models.DateTimeField()
    user_id = models.IntegerField()
    content_type_id = models.IntegerField()
    object_id = models.TextField()
    object_repr = models.CharField(max_length=200)
    action_flag = models.SmallIntegerField()
    change_message = models.TextField()
    class Meta:
        db_table = u'django_admin_log'

class DjangoContentType(models.Model):
    id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=100)
    app_label = models.CharField(max_length=100)
    model = models.CharField(max_length=100)
    class Meta:
        db_table = u'django_content_type'

class DjangoSession(models.Model):
    session_key = models.CharField(max_length=40, primary_key=True)
    session_data = models.TextField()
    expire_date = models.DateTimeField()
    class Meta:
        db_table = u'django_session'

class DjangoSite(models.Model):
    id = models.IntegerField(primary_key=True)
    domain = models.CharField(max_length=100)
    name = models.CharField(max_length=50)
    class Meta:
        db_table = u'django_site'

class GeographyColumns(models.Model):
    f_table_catalog = models.TextField() # This field type is a guess.
    f_table_schema = models.TextField() # This field type is a guess.
    f_table_name = models.TextField() # This field type is a guess.
    f_geography_column = models.TextField() # This field type is a guess.
    coord_dimension = models.IntegerField()
    srid = models.IntegerField()
    type = models.TextField()
    class Meta:
        db_table = u'geography_columns'

class GeometryColumns(models.Model):
    f_table_catalog = models.CharField(max_length=256)
    f_table_schema = models.CharField(max_length=256)
    f_table_name = models.CharField(max_length=256)
    f_geometry_column = models.CharField(max_length=256)
    coord_dimension = models.IntegerField()
    srid = models.IntegerField()
    type = models.CharField(max_length=30)
    class Meta:
        db_table = u'geometry_columns'

class RasterColumns(models.Model):
    r_table_catalog = models.TextField() # This field type is a guess.
    r_table_schema = models.TextField() # This field type is a guess.
    r_table_name = models.TextField() # This field type is a guess.
    r_raster_column = models.TextField() # This field type is a guess.
    srid = models.IntegerField()
    scale_x = models.FloatField()
    scale_y = models.FloatField()
    blocksize_x = models.IntegerField()
    blocksize_y = models.IntegerField()
    same_alignment = models.BooleanField()
    regular_blocking = models.BooleanField()
    num_bands = models.IntegerField()
    pixel_types = models.TextField() # This field type is a guess.
    nodata_values = models.TextField() # This field type is a guess.
    out_db = models.TextField() # This field type is a guess.
    extent = models.GeometryField(srid=0)
    objects = models.GeoManager()
    class Meta:
        db_table = u'raster_columns'

class RasterOverviews(models.Model):
    o_table_catalog = models.TextField() # This field type is a guess.
    o_table_schema = models.TextField() # This field type is a guess.
    o_table_name = models.TextField() # This field type is a guess.
    o_raster_column = models.TextField() # This field type is a guess.
    r_table_catalog = models.TextField() # This field type is a guess.
    r_table_schema = models.TextField() # This field type is a guess.
    r_table_name = models.TextField() # This field type is a guess.
    r_raster_column = models.TextField() # This field type is a guess.
    overview_factor = models.IntegerField()
    class Meta:
        db_table = u'raster_overviews'

class SpatialRefSys(models.Model):
    srid = models.IntegerField(primary_key=True)
    auth_name = models.CharField(max_length=256)
    auth_srid = models.IntegerField()
    srtext = models.CharField(max_length=2048)
    proj4text = models.CharField(max_length=2048)
    class Meta:
        db_table = u'spatial_ref_sys'

class VwConcepts(models.Model):
    conceptid = models.TextField(primary_key=True) # This field type is a guess.
    conceptlabel = models.TextField()
    legacyoid = models.TextField()
    class Meta:
        db_table = u'vw_concepts'

class VwEdges(models.Model):
    source = models.TextField() # This field type is a guess.
    label = models.TextField()
    target = models.TextField() # This field type is a guess.
    class Meta:
        db_table = u'vw_edges'

class VwEntityTypes(models.Model):
    entitytypeid = models.TextField()
    entitytypename = models.TextField()
    description = models.TextField()
    languageid = models.TextField()
    reportwidget = models.TextField()
    icon = models.TextField()
    isresource = models.BooleanField()
    conceptid = models.TextField() # This field type is a guess.
    class Meta:
        db_table = u'vw_entity_types'

class VwNodes(models.Model):
    label = models.TextField()
    id = models.TextField(primary_key=True) # This field type is a guess.
    val = models.TextField()
    class Meta:
        db_table = u'vw_nodes'

class VwEntitytypeDomains(models.Model):
    id = models.TextField(primary_key=True) # This field type is a guess.
    entitytypeid = models.TextField(blank=True)
    conceptid = models.TextField(blank=True) # This field type is a guess.
    value = models.TextField(blank=True)
    valuetype = models.TextField(blank=True)
    languageid = models.TextField(blank=True)
    sortorder = models.TextField(blank=True)
    class Meta:
        managed = False
        db_table = 'vw_entitytype_domains'

class AuthMessage(models.Model):
    id = models.IntegerField(primary_key=True)
    user = models.ForeignKey('AuthUser')
    message = models.TextField()
    class Meta:
        db_table = u'auth_message'

class AuthPermission(models.Model):
    id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=50)
    content_type = models.ForeignKey('DjangoContentType')
    codename = models.CharField(max_length=100)
    class Meta:
        db_table = u'auth_permission'

class Concepts(models.Model):
    conceptid = models.TextField(primary_key=True) # This field type is a guess.
    nodetype = models.ForeignKey('DNodetypes', db_column='nodetype')
    legacyoid = models.TextField()
    class Meta:
        db_table = u'concepts'

    def __unicode__(self):
        return ('%s') % (self.conceptid)

class ConceptRelations(models.Model):
    relationid = models.TextField(primary_key=True)
    conceptidfrom = models.ForeignKey('Concepts', db_column='conceptidfrom', related_name='concept_relation_from')
    conceptidto = models.ForeignKey('Concepts', db_column='conceptidto', related_name='concept_relation_to')
    relationtype = models.ForeignKey('DRelationtypes', db_column='relationtype', related_name='concept_relation_type')
    class Meta:
        db_table = u'concepts"."relations'

    def __unicode__(self):
        return ('%s') % (self.relationid)

    def save(self, *args, **kwargs):
        # prevent insert of duplicate records
        if ConceptRelations.objects.filter(conceptidfrom = self.conceptidfrom, conceptidto = self.conceptidto, relationtype = self.relationtype).exists():
            return # can't insert duplicate values
        elif self.relationtype_id == 'related':
            # check to see if the recipocal relationship exists as well (mostly for when we load a skos file)
            if ConceptRelations.objects.filter(conceptidfrom = self.conceptidto, conceptidto = self.conceptidfrom, relationtype = self.relationtype).exists():
                return

        super(ConceptRelations, self).save(*args, **kwargs) # Call the "real" save() method.

class ValueTypes(models.Model):
    valuetype = models.TextField(primary_key=True)
    category = models.TextField()
    description = models.TextField()
    namespace = models.TextField()
    class Meta:
        db_table = u'concepts"."d_valuetypes'

    def __unicode__(self):
        return ('valuetype: %s, category: %s') % (self.valuetype, self.category)

class Values(models.Model):
    valueid = models.TextField(primary_key=True) # This field type is a guess.
    conceptid = models.ForeignKey('Concepts', db_column='conceptid')
    valuetype = models.ForeignKey('ValueTypes', db_column='valuetype')
    value = models.TextField()
    languageid = models.ForeignKey('DLanguages', db_column='languageid')
    class Meta:
        db_table = u'values'

class FileValues(models.Model):
    valueid = models.TextField(primary_key=True) # This field type is a guess.
    conceptid = models.ForeignKey('Concepts', db_column='conceptid')
    valuetype = models.ForeignKey('ValueTypes', db_column='valuetype')
    value = models.FileField(upload_to='concepts')
    languageid = models.ForeignKey('DLanguages', db_column='languageid')
    class Meta:
        db_table = u'values'

    def geturl(self):
        if self.value != None:
            return self.value.url
        return ''

    def getname(self):
        if self.value != None:
            return self.value.name[6:]
        return ''

class AuthGroupPermissions(models.Model):
    id = models.IntegerField(primary_key=True)
    group = models.ForeignKey('AuthGroup')
    permission = models.ForeignKey('AuthPermission')
    class Meta:
        db_table = u'auth_group_permissions'

class AuthUserGroups(models.Model):
    id = models.IntegerField(primary_key=True)
    user = models.ForeignKey('AuthUser')
    group = models.ForeignKey('AuthGroup')
    class Meta:
        db_table = u'auth_user_groups'

class AuthUserUserPermissions(models.Model):
    id = models.IntegerField(primary_key=True)
    user = models.ForeignKey('AuthUser')
    permission = models.ForeignKey('AuthPermission')
    class Meta:
        db_table = u'auth_user_user_permissions'

class Properties(models.Model):
    propertyid = models.TextField(primary_key=True)
    classdomain = models.ForeignKey('Classes', db_column='classdomain', related_name='properties_classdomain')
    classrange = models.ForeignKey('Classes', db_column='classrange', related_name='properties_classrange')
    propertydisplay = models.TextField()
    class Meta:
        db_table = u'properties'

class EntityTypes(models.Model):
    classid = models.ForeignKey('Classes', db_column='classid', related_name='entitytypes_classid')
    conceptid = models.ForeignKey('Concepts', db_column='conceptid', related_name='entitytypes_conceptid')
    businesstablename = models.TextField()
    publishbydefault = models.BooleanField()
    icon = models.TextField()
    defaultvectorcolor = models.TextField()
    entitytypeid = models.TextField(primary_key=True)
    isresource = models.BooleanField()
    class Meta:
        db_table = u'data"."entity_types'

    def getcolumnname(self):
        ret = None
        if self.businesstablename is not None and self.businesstablename != 'entities':
            ret = 'val'

        return ret     

    def __unicode__(self):
        return ('%s') % (self.entitytypeid)

class Entities(models.Model):
    entityid = models.TextField(primary_key=True) # This field type is a guess.
    entitytypeid = models.ForeignKey('EntityTypes', db_column='entitytypeid')
    class Meta:
        db_table = u'data"."entities'

    def __unicode__(self):
        return ('%s of type %s') % (self.entityid, self.entitytypeid)

class Strings(models.Model):
    entityid = models.ForeignKey('Entities', primary_key=True, db_column='entityid')
    val = models.TextField()
    class Meta:
        db_table = u'strings'

class Dates(models.Model):
    entityid = models.ForeignKey('Entities', primary_key=True, db_column='entityid')
    val = models.DateTimeField()
    class Meta:
        db_table = u'dates'

    def __unicode__(self):
        return ('%s') % (self.val)

class Numbers(models.Model):
    entityid = models.ForeignKey('Entities', primary_key=True, db_column='entityid')
    val = models.DecimalField(max_digits=65535, decimal_places=65534)
    class Meta:
        db_table = u'numbers'
        
class Files(models.Model):
    entityid = models.ForeignKey('Entities', primary_key=True, db_column='entityid')
    val = models.FileField(upload_to='files')
    class Meta:
        db_table = u'files'

    def geturl(self):
        if self.val != None:
            return self.val.url
        return ''

    def getname(self):
        if self.val != None:
            return self.val.name[6:]
        return ''

# These two event listeners auto-delete files from filesystem when they are unneeded:
# from http://stackoverflow.com/questions/16041232/django-delete-filefield
@receiver(models.signals.post_delete, sender=Files)
def auto_delete_file_on_delete(sender, instance, **kwargs):
    """Deletes file from filesystem
    when corresponding `Files` object is deleted.
    """
    if instance.val:
        if os.path.isfile(instance.val.path):
            os.remove(instance.val.path)

@receiver(models.signals.pre_save, sender=Files)
def auto_delete_file_on_change(sender, instance, **kwargs):
    """Deletes file from filesystem
    when corresponding `Files` object is changed.
    """

    if not instance.pk:
        return False

    try:
        old_file = Files.objects.get(pk=instance.pk).val
    except Files.DoesNotExist:
        return False

    new_file = instance.val
    if not old_file == new_file:
        try:
            if os.path.isfile(old_file.path):
                os.remove(old_file.path)
        except Exception:
            return False

class Geometries(models.Model):
    entityid = models.ForeignKey('Entities', primary_key=True, db_column='entityid')
    val = models.GeometryField()
    objects = models.GeoManager()
    class Meta:
        db_table = u'geometries'

class EditLog(models.Model):
    editlogid = models.TextField(primary_key=True)
    resourceentitytypeid = models.TextField()
    resourceid = models.TextField()
    attributeentitytypeid = models.TextField()
    edittype = models.TextField()
    userid = models.TextField()
    timestamp = models.DateTimeField()
    oldvalue = models.TextField()
    newvalue = models.TextField()
    user_email = models.TextField()
    user_firstname = models.TextField()
    user_lastname = models.TextField()
    note = models.TextField()
    class Meta:
        db_table = u'data"."edit_log'

class Domains(models.Model):
    entityid = models.ForeignKey('Entities', primary_key=True, db_column='entityid')
    val = models.ForeignKey('Values', db_column='val', null=True)
    class Meta:
        db_table = u'domains'

    def getlabelid(self):
        if self.val_id != None:
            return self.val_id
        return ''

    def getlabelvalue(self):
        if self.val != None:
            return self.val.value   
        return ''

    # def __getattr__(self, name):
    #     print name        
    #     if name == 'val' and self.val_id != None:

    #         return self.val_id
    #     else:
    #         return super(Domains, self).__getattr__(name)

    def __setattr__(self, name, value):
        if name == 'val':
            if not isinstance(value, Values):
                try:
                    uuid.UUID(value)
                    value = Values.objects.get(valueid = value)
                except(ValueError, TypeError):
                    # print value, 'attr value'
                    concepts = Concepts.objects.filter(legacyoid = value)
                    if len(concepts) == 1:
                        value = Values.objects.get(conceptid = concepts[0], valuetype = 'prefLabel')
                    else:
                        # print 'unable to find, or found more then 1 Concept with legacyoid: %s' % (value)
                        value = None

        super(Domains, self).__setattr__(name, value)

class Mappings(models.Model):
    mappingid = models.TextField(primary_key=True) # This field type is a guess.
    entitytypeidfrom = models.ForeignKey('EntityTypes', db_column='entitytypeidfrom', related_name='mappings_entitytypeidfrom')
    entitytypeidto = models.ForeignKey('EntityTypes', db_column='entitytypeidto', related_name='mappings_entitytypeidto')
    mergenodeid = models.TextField()
    class Meta:
        db_table = u'ontology"."mappings'

    def __unicode__(self):
        return ('%s -- %s') % (self.entitytypeidfrom, self.entitytypeidto)

class Rules(models.Model):
    ruleid = models.TextField(primary_key=True) # This field type is a guess.
    entitytypedomain = models.ForeignKey('EntityTypes', db_column='entitytypedomain', related_name='rules_entitytypedomain')
    entitytyperange = models.ForeignKey('EntityTypes', db_column='entitytyperange', related_name='rules_entitytyperange')
    propertyid = models.ForeignKey('Properties', db_column='propertyid')
    class Meta:
        db_table = u'rules'

class MappingSteps(models.Model):
    mappingid = models.ForeignKey('Mappings', primary_key=True, db_column='mappingid')
    ruleid = models.ForeignKey('Rules', db_column='ruleid')
    order = models.IntegerField()
    class Meta:
        db_table = u'mapping_steps'

class Relations(models.Model):
    relationid = models.AutoField(primary_key=True)
    ruleid = models.ForeignKey('Rules', db_column='ruleid')
    entityiddomain = models.ForeignKey('Entities', db_column='entityiddomain', related_name='rules_entityiddomain')
    entityidrange = models.ForeignKey('Entities', db_column='entityidrange', related_name='rules_entityidrange')
    notes = models.TextField()
    class Meta:
        db_table = u'data"."relations'

    def __unicode__(self):
        return ('%s is the parent of %s') % (self.entityiddomain, self.entityidrange)

    def save(self, *args, **kwargs):
        if Relations.objects.filter(ruleid = self.ruleid, entityiddomain = self.entityiddomain, entityidrange = self.entityidrange).exists():
            return # can't insert duplicate values
        else:
            super(Relations, self).save(*args, **kwargs) # Call the "real" save() method.

class RelatedResource(models.Model):
    resourcexid = models.AutoField(primary_key=True)
    entityid1 = models.TextField()
    entityid2 = models.TextField()
    notes = models.TextField()
    relationshiptype = models.TextField()
    datestarted = models.DateField()
    dateended = models.DateField()
    class Meta:
        db_table = u'data"."resource_x_resource'

class UserProfile(models.Model):
    user = models.OneToOneField(User, primary_key=True)
    details = models.TextField()

    class Meta:
        db_table = u'user_profile'

    def __unicode__(self):
        return str(self.user.username)

    def save(self, *args, **kwargs):
        """Validates JSON prior to saving"""
        try:
            json.loads(self.details)
            super(UserProfile, self).save(*args, **kwargs) # Call the "real" save() method.
        except:
            return #'json invalid'


import sys
import json
# from .. import utils
from django.contrib.gis.gdal import DataSource
from django.conf import settings
import os
from arches.app.models.models import Concepts
from arches.app.models.models import VwConcepts
from arches.app.models.models import EntityTypes
from arches.app.models.concept import Concept
from django.db import connection
import arches.management.commands.utils as utils


class Row(object):
    def __init__(self, *args):
        if len(args) == 0:
            self.resource_id = ''
            self.resourcetype = ''
            self.attributename = ''
            self.attributevalue = ''
            self.group_id = ''
        elif isinstance(args[0], list):
            self.resource_id = args[0][0].strip()
            self.resourcetype = args[0][1].strip()
            self.attributename = args[0][2].strip()
            self.attributevalue = args[0][3].strip().replace('\\n', '\n')
            self.group_id = args[0][4].strip()

    def __repr__(self):
        return ('"%s, %s, %s, %s"') % (self.resource_id, self. resourcetype, self. attributename, self.attributevalue)


class Group(object):
    def __init__(self, *args):
        if len(args) == 0:
            self.resource_id = ''
            self.group_id = ''
            self.rows = []
        elif isinstance(args[0], list):
            self.resource_id = args[0][0].strip()
            self.group_id = args[0][4].strip()
            self.rows = []   


class Resource(object):

    def __init__(self, *args):

        if len(args) == 0:
            self.resource_id = ''
            self.entitytypeid = ''
            self.groups = []
            self.nongroups = []

        elif isinstance(args[0], list):
            self.entitytypeid = args[0][1].strip()
            self.resource_id = args[0][0].strip()
            self.nongroups = []  
            self.groups = []


    def appendrow(self, row, group_id=None):
        if group_id != None:
            for group in self.groups:
                if group.group_id == group_id:
                    group.rows.append(row)
        else:
           self.nongroups.append(row)

    def __str__(self):
        return '{0},{1}'.format(self.resource_id, self.entitytypeid)


class ShapeReader():

    def load_file(self, shapefile):
        resources = []
        self.configs = self.parse_configs(shapefile)
        if self.configs:
            self.attr_map = dict(self.configs['FIELD_MAP'])

            self.concept_value_mappings = {}
            for field, entitytypeid in self.attr_map.iteritems():
                if entitytypeid.endswith('.E55'):
                    key, mapping = self.get_e55_concept_legacyoids(entitytypeid)
                    self.concept_value_mappings[key] = mapping
                else:
                    pass      
            
            self.auth_map = self.convert_uuid_map_to_conceptid_map(self.convert_aux_map_to_uuid_map(self.configs['AUXILIARY_MAP']))
            self.entitytypeid = self.configs['RESOURCE_TYPE']
            self.geom_type = self.configs['GEOM_TYPE']

            self.shp_data = self.read_shapefile(shapefile)

            shp_resource_info = self.collect_resource_info(
                    attr_mapping=self.attr_map,
                    auth_mapping=self.auth_map,
                    reader_output=self.shp_data,
                    geom_type=self.geom_type,
                    value_to_concept_label_mappings=self.concept_value_mappings)

            resource_id = ''
            group_id = ''
                
            for shp_dictionary in shp_resource_info:
                
                if (settings.LIMIT_ENTITY_TYPES_TO_LOAD == None or self.entitytypeid in settings.LIMIT_ENTITY_TYPES_TO_LOAD):
                    resource = Resource()
                    for key in shp_dictionary.keys():
                        if key is not None:
                            row = Row()
                            row.resource_id = resource_id
                            row.resourcetype = self.entitytypeid
                            row.attributename = key
                            row.attributevalue = shp_dictionary[key]
                            resource.appendrow(row)
                    resource.entitytypeid = self.entitytypeid
                    resources.append(resource)

        return resources


    def read(self, path):
        '''
        return the read shapefile
        '''
        shp_file = DataSource(os.path.abspath(path))
        return shp_file
        

    def get_layer_details(self, shp_file):
        '''
        Returns layer details of the given shapefile.
        Note that, we return a list even though 
            shapefiles are only allowed to have one layer
        '''
        layer_list = []
        for layer in shp_file:
            layer_list.append(layer)
        return layer_list
    

    def get_field_names(self, shp_file):
        '''
        return available fields of the passed shp file 
        '''
        fields = []
        # we get fields for all the layers
        layers = self.get_layer_details(shp_file)
        for layer in layers:
            fields.append(layer.fields)
        # returns a list of lists
        return fields
    

    def get_values(self, shp_file):
        '''
        returns a list of data associated with each field.
        field_values[n] is a list containing all the values for the 'n'th attribute
        '''
        field_names = self.get_field_names(shp_file)
        layer_0 = self.get_layer_details(shp_file)[0]
        field_values = []
    
        for field in field_names[0]: #consider the 0th layer
            field_values.append(layer_0.get_fields(field))
        return field_values
    

    def get_wkt_values(self, shp_file):
        '''
        returns geographic data associated with the shapefile(shapes) in WKT format.
        These value will be added to the json file 
        '''
        wkt_list=[]
        #return geo_data in wkt format
        for resource in self.get_layer_details(shp_file)[0]:
            wkt_list.append(resource.geom.wkt)
        return wkt_list


    def parse_configs(self, shapefile):
        '''
        Takes a shapefile and reads a config file with the same basename.
        Returns a dictionary with configuration for creating resources from shapefile records
        '''
        result = None
        try:
            config_file = os.path.join(os.path.dirname(shapefile), os.path.basename(shapefile).split('.')[0] + '.config')
            if os.path.exists(config_file):
                result = json.load(open(config_file, 'r'))
                result['AUXILIARY_MAP'] = {}
            else:
                print 'Unable to find config file:', config_file
            #AUXILIARY_MAP = {"NAME TYPE.E55" : "Primary","EXTERNAL XREF TYPE.E55" : "Legacy System"}
            #The auxiliary_map was Tharindus strategy to map concepts to their labels. This is not being implemented yet
        except:
            print "config file is missing or improperly named. Make sure you have config file with the same basename as your shapefile and the extension .config"

        return result


    def get_concept_uuid_for_aux_map_entry(self,concept_name,concept_value):
        '''
        Reads concept labels(prefLabels) given in AUXILIARY_MAP and returns a list of conceptids.

        '''
        concept = None
        try:
            concept = Concepts.objects.get(legacyoid = concept_name)
        except:
            print "No Concept found with the name %s"%concept_name
        if concept is not None:
            concept_graph = Concept().get(id=concept.pk, include_subconcepts=True, include=['label'])
   
            if concept_graph.subconcepts:
                for subconcept  in concept_graph.subconcepts[0].subconcepts:
                    for value in subconcept.values:

                        if value.type == "prefLabel" and value.value == concept_value:
                            return subconcept.id

                print '[ERROR]: %s is not found in %s'%(concept_value,concept_name)
            return
        

    def convert_aux_map_to_uuid_map(self,aux_map):
        '''
        converts the AUX_MAP values into uuid
        '''
        converted_aux_map = {}
        for entry in aux_map.keys():
            value = self.get_concept_uuid_for_aux_map_entry(entry, aux_map[entry])
            if value is not None:
                converted_aux_map[entry] = value
        
        return converted_aux_map  
    

    def convert_uuid_map_to_conceptid_map(self,uuid_map):
        '''
        convert an uuid dictionary into conceptId dictionary because
        data.domains should be populated with conceptId values
        ''' 
        conceptid_map = {}
        sql_template = """'SELECT concepts.legacyoid FROM concepts.concepts WHERE concepts.conceptid = '"""
        cursor = connection.cursor()
        for key in uuid_map.keys():
            cursor.execute(sql_template+uuid_map[key]+"'")
            conceptid_map[key] = cursor.fetchone()[0]
            
        return conceptid_map  


    def get_e55_concept_legacyoids(self, e55_type, lang=u'en-us'):
        concept = EntityTypes.objects.get(pk=e55_type).conceptid
        concept_graph = Concept().get(id=concept.pk, include_relatedconcepts=True, include=['label'])
        values_to_legacy = []
        cursor = connection.cursor()
        if len(concept_graph.relatedconcepts) > 0:
            for concept_graph in concept_graph.relatedconcepts:
                for value in concept_graph.values:
                    if value.type == "prefLabel" and value.language == lang:
                        concept_view = VwConcepts.objects.filter(conceptid = value.conceptid)[0] 
                        values_to_legacy.append({value.value:concept_view.legacyoid})
        return e55_type, values_to_legacy


    # Now build a list of dictionaries, one per record
    def collect_resource_info(self, attr_mapping, auth_mapping, reader_output, geom_type, value_to_concept_label_mappings, break_on_error=True):

        dict_list=[] # list of dictionaries
        attr_names = reader_output[0]
        attr_vals = reader_output[1][0:-1] # get all attribute values except the geom_wkt values
        geom_values = reader_output[1][-1] # last index because we append wkt values at the end

        '''
        first, add the attribute values to the dictionary
        and then add authority details. Because shapefile data does not 
        contain authority details, the authority mapping
        is defined by the user and passed in a separate dictionary
        '''      
        errors = []
        for record_index in range (0, len(attr_vals[0])): #len(attr_vals[0]) equals to the number of records
            record_dictionary= {} # i th dictionary
            for attr in attr_names[0]: # loops for (number of attributes) times
                #get the index of the selected attribute
                attr_index = attr_names[0].index(attr)
                #add the key/value pair retrieved from the attr_mapping
                entitytypeid = attr_mapping.get(attr)
                label = attr_vals[attr_index][record_index]
                found_match = False
                if type(entitytypeid) in [str, unicode]:
                    if entitytypeid.endswith('.E55'):
                        count = 0
                        for mapping in value_to_concept_label_mappings[entitytypeid]:
                            count += 1
                            if unicode(label) == mapping.keys()[0]:
                                label = mapping[label]
                                count = 0
                                break
                        if count == len(value_to_concept_label_mappings[entitytypeid]):
                            errors.append('shapefile record {0}: "{1}", Does not match any available {2} concept value\n'.format(str(record_index), label, entitytypeid))
                        else:
                            print str(record_index), label, entitytypeid
                        
                record_dictionary[entitytypeid] = label
            
            record_dictionary.update(auth_mapping)

            record_dictionary[geom_type] = geom_values[record_index] 
            dict_list.append(record_dictionary)

        if len(errors) > 0:
            utils.write_to_file(os.path.join(settings.PACKAGE_ROOT, 'logs', 'shapefile_loading_errors.txt'), '\n'.join(errors))
            print 'There were errors matching some values to concepts, please see {0} for details'.format(os.path.join(settings.PACKAGE_ROOT, 'logs', 'shapefile_loading_errors.txt'))
            if break_on_error:
                sys.exit(101)

        return dict_list


    def read_shapefile(self, shp_path):
        '''
        The following method takes the location of a shapefile returns all data as a single tuple.
        '''

        shapefile = self.read(shp_path)
        field_names = self.get_field_names(shapefile)
        field_values_and_geom_wkt = self.get_values(shapefile)
        geom_wkt_values = self.get_wkt_values(shapefile)
        
        # since it is desirable to pass field_values and wkt data in a single list, we combine the two lists
        field_values_and_geom_wkt.append(geom_wkt_values) #the last item of field_values is a list containing wkt geom values
        
        return (field_names, field_values_and_geom_wkt)



'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

from django.conf.urls import patterns, include, url
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.conf.urls.i18n import patterns

# Uncomment the next two lines to enable the admin:
from django.contrib import admin
admin.autodiscover()

uuid_regex = '[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}'

urlpatterns = patterns('',
    url(r'^$', 'arches.app.views.main.index'),
    url(r'^index.htm', 'arches.app.views.main.index', name='home'),
    url(r'^auth/', 'arches.app.views.main.auth', name='auth'),
    url(r'^rdm/(?P<conceptid>%s|())$' % uuid_regex , 'arches.app.views.concept.rdm', name='rdm'),
    url(r'^map', 'arches.app.views.map.get_page', name="map"),
    url(r'^geocoder', 'arches.app.views.search.geocode', name="geocoder"),
    
    url(r'^entities/(?P<entityid>%s)$' % uuid_regex , 'arches.app.views.entity.Entities'),
    url(r'^entityTypes/(?P<entitytypeid>.*)$', 'arches.app.views.entity.EntityTypes'),
    url(r'^concepts/(?P<conceptid>%s)/manage_parents/$' % uuid_regex, 'arches.app.views.concept.manage_parents', name="concept_manage_parents"),        
    url(r'^concepts/(?P<conceptid>%s)/confirm_delete/$' % uuid_regex, 'arches.app.views.concept.confirm_delete', name="confirm_delete"),     
    url(r'^concepts/(?P<conceptid>%s|())/$' % uuid_regex , 'arches.app.views.concept.concept', name="concept"),
    url(r'^concepts/tree', 'arches.app.views.concept.concept_tree', name="concept_tree"),      
    url(r'^concepts/search', 'arches.app.views.concept.search', name="concept_search"),
    url(r'^search$', 'arches.app.views.search.home_page', name="search_home"),
    url(r'^search/terms$', 'arches.app.views.search.search_terms', name="search_terms"),
    url(r'^search/resources$', 'arches.app.views.search.search_results', name="search_results"),
    url(r'^search/export$', 'arches.app.views.search.export_results', name="search_results_export"),
    url(r'^buffer/$', 'arches.app.views.search.buffer', name="buffer"),
    url(r'^resources/(?P<resourcetypeid>[0-9a-zA-Z_.]*)/(?P<form_id>[a-zA-Z_-]*)/(?P<resourceid>%s|())$' % uuid_regex, 'arches.app.views.resources.resource_manager', name="resource_manager"),
    url(r'^resources/related/(?P<resourceid>%s|())$' % uuid_regex, 'arches.app.views.resources.related_resources', name="related_resources"),
    url(r'^resources/history/(?P<resourceid>%s|())$' % uuid_regex, 'arches.app.views.resources.edit_history', name="edit_history"),
    url(r'^resources/layers/(?P<entitytypeid>.*)$', 'arches.app.views.resources.map_layers', name="map_layers"),
    url(r'^resources/markers/(?P<entitytypeid>.*)$', 'arches.app.views.resources.map_layers', {'get_centroids':True}, name="map_markers"),

    # Uncomment the admin/doc line below to enable admin documentation:
    # url(r'^admin/doc/', include('django.contrib.admindocs.urls')),

    # Uncomment the next line to enable the admin:
    url(r'^admin/', include(admin.site.urls)),
)

urlpatterns += staticfiles_urlpatterns()



'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

import re
from django.template import RequestContext
from django.shortcuts import render_to_response, redirect
from django.views.decorators.csrf import csrf_exempt
from django.contrib.auth.decorators import login_required
from django.conf import settings
from django.db import transaction
from arches.app.models import models
from arches.app.models.resource import Resource
from arches.app.utils.betterJSONSerializer import JSONSerializer, JSONDeserializer
from arches.app.utils.JSONResponse import JSONResponse
from arches.app.models.entity import Entity
from arches.app.search.search_engine_factory import SearchEngineFactory
from arches.app.search.elasticsearch_dsl_builder import Query, Terms
from arches.app.views.concept import get_preflabel_from_valueid

@login_required
@csrf_exempt
def resource_manager(request, resourcetypeid='', form_id='', resourceid=''):

    if resourceid != '':
        resource = Resource(resourceid)
    elif resourcetypeid != '':
        resource = Resource({'entitytypeid': resourcetypeid})

    form = resource.get_form(form_id)

    if request.method == 'POST':
        # get the values from the form and pass to the resource
        data = JSONDeserializer().deserialize(request.POST.get('formdata', {}))
        form.update(data)

        with transaction.atomic():
            if resourceid != '':
                resource.delete_index()
            resource.save(user=request.user)
            resource.index()
            resourceid = resource.entityid

            return redirect('resource_manager', resourcetypeid=resourcetypeid, form_id=form_id, resourceid=resourceid)

    return render_to_response('resource-manager.htm', {
            'form': form,
            'formdata': JSONSerializer().serialize(form.data),
            'form_template': 'views/forms/' + form_id + '.htm',
            'form_id': form_id,
            'resourcetypeid': resourcetypeid,
            'resourceid': resourceid,
            'main_script': 'resource-manager',
            'active_page': 'ResourceManger',
            'resource': resource,
            'resource_name': resource.get_primary_name(),
            'resource_type_name': resource.get_type_name(),
            'form_groups': resource.form_groups
        },
        context_instance=RequestContext(request))
    
def related_resources(request, resourceid):
    lang = request.GET.get('lang', settings.LANGUAGE_CODE)
    start = request.GET.get('start', 0)
    return JSONResponse(get_related_resources(resourceid, lang, start=start, limit=15), indent=4)

def get_related_resources(resourceid, lang, limit=1000, start=0):
    ret = {
        'resource_relationships': [],
        'related_resources': []
    }
    se = SearchEngineFactory().create()

    query = Query(se, limit=limit, start=start)
    query.add_filter(Terms(field='entityid1', terms=resourceid).dsl, operator='or')
    query.add_filter(Terms(field='entityid2', terms=resourceid).dsl, operator='or')
    resource_relations = query.search(index='resource_relations', doc_type='all')
    ret['total'] = resource_relations['hits']['total']

    entityids = set()
    for relation in resource_relations['hits']['hits']:
        relation['_source']['preflabel'] = get_preflabel_from_valueid(relation['_source']['relationshiptype'], lang)
        ret['resource_relationships'].append(relation['_source'])
        entityids.add(relation['_source']['entityid1'])
        entityids.add(relation['_source']['entityid2'])
    if len(entityids) > 0:
        entityids.remove(resourceid)   

    related_resources = se.search(index='entity', doc_type='_all', id=list(entityids))
    if related_resources:
        for resource in related_resources['docs']:
            ret['related_resources'].append(resource['_source'])

    return ret

def map_layers(request, entitytypeid='all', get_centroids=False):
    data = []
    bbox = request.GET.get('bbox', '')
    limit = request.GET.get('limit', settings.MAP_LAYER_FEATURE_LIMIT)
    entityids = request.GET.get('entityid', '')
    geojson_collection = {
      "type": "FeatureCollection",
      "features": []
    }
    
    se = SearchEngineFactory().create()
    query = Query(se, limit=limit)

    args = { 'index': 'maplayers' }
    if entitytypeid != 'all':
        args['doc_type'] = entitytypeid
    if entityids != '':
        for entityid in entityids.split(','):
            geojson_collection['features'].append(se.search(index='maplayers', id=entityid)['_source'])
        return JSONResponse(geojson_collection)

    data = query.search(**args)

    for item in data['hits']['hits']:
        if get_centroids:
            item['_source']['geometry'] = item['_source']['properties']['centroid']
            item['_source'].pop('properties', None)
        else:
            item['_source']['properties'].pop('extent', None)
            item['_source']['properties'].pop('centroid', None)
        geojson_collection['features'].append(item['_source'])

    return JSONResponse(geojson_collection)

def edit_history(request, resourceid=''):
    ret = []
    current = None
    index = -1
    start = request.GET.get('start', 0)
    limit = request.GET.get('limit', 10)
    if resourceid != '':
        dates = models.EditLog.objects.filter(resourceid = resourceid).values_list('timestamp', flat=True).order_by('-timestamp').distinct('timestamp')[start:limit]
        # dates = models.EditLog.objects.datetimes('timestamp', 'second', order='DESC')
        for date in dates:
            #ret[str(date)] = models.EditLog.objects.filter(resourceid = self.resource.entityid, timestamp = date)
            print str(date)

        for log in models.EditLog.objects.filter(resourceid = resourceid, timestamp__in = dates).values().order_by('-timestamp', 'attributeentitytypeid'):
            if str(log['timestamp']) != current:
                current = str(log['timestamp']) 
                ret.append({'date':str(log['timestamp'].date()), 'time': str(log['timestamp'].time().replace(microsecond=0).isoformat()), 'log': []})
                index = index + 1

            ret[index]['log'].append(log)
            
    return JSONResponse(ret, indent=4)


class ResourceForm(object):
    def __init__(self, resource):
        # here is where we can create the basic format for the form data
        info = self.get_info()
        self.id = info['id']
        self.name = info['name']
        self.icon = info['icon']
        self.resource = resource
        if self.resource == None:
            self.schema = None
        else:
            self.schema = Entity.get_mapping_schema(self.resource.entitytypeid)
        
        self.data = {
            "domains": {},
            "defaults": {}
        }
        self.load()
    
    @staticmethod
    def get_info():
        return {
            'id': '',
            'icon': '',
            'name': '',
            'class': ResourceForm
        }

	def update(self, data):
		# update resource w/ post data
		return 

    def load(self):
        # retrieves the data from the server
        return 

    def get_nodes(self, entitytypeid):
        ret = []
        entities = self.resource.find_entities_by_type_id(entitytypeid)
        for entity in entities:
            data = {}

            for entity in entity.flatten():
                #data[entity.entitytypeid] = self.encode_entity(entity)
                data = dict(data.items() + self.encode_entity(entity).items())
            ret.append(data)
        return ret

    def update_nodes(self, entitytypeid, data):
        for entity in self.resource.find_entities_by_type_id(entitytypeid):
            self.resource.child_entities.remove(entity)

        if self.schema == None:
            self.schema = Entity.get_mapping_schema(self.resource.entitytypeid)
        for value in data[entitytypeid.replace('.', '_')]:
            baseentity = None
            for newentity in self.decode_data_item(value):
                entity = Entity()
                entity.create_from_mapping(self.resource.entitytypeid, self.schema[newentity['entitytypeid']]['steps'], newentity['entitytypeid'], newentity['value'], newentity['entityid'])

                if baseentity == None:
                    baseentity = entity
                else:
                    baseentity.merge(entity)
            
            self.resource.merge_at(baseentity, self.resource.entitytypeid)

    def update_node(self, entitytypeid, data):
        if self.schema == None:
            self.schema = Entity.get_mapping_schema(self.resource.entitytypeid)
        nodes = self.resource.find_entities_by_type_id(entitytypeid)
        node_data = self.decode_data_item(data[entitytypeid.replace('.', '_')])[0]

        if len(nodes) == 0:
            entity = Entity()
            entity.create_from_mapping(self.resource.entitytypeid, self.schema[entitytypeid]['steps'], entitytypeid, node_data['value'], node_data['entityid'])
            self.resource.merge_at(entity, self.resource.entitytypeid)
        else:
            nodes[0].value = node_data['value']

    def encode_entity(self, entity):
        def enc(entity, attr):
            return '%s__%s' % (entity.entitytypeid.replace('.', '_'), attr)

        ret = {}
        for key, value in entity.__dict__.items():
            if not key.startswith("__"):
                ret[enc(entity, key)] = value
        return ret

    def decode_data_item(self, entity):
        def dec(item):
            # item = "NAME_E41__entitytypeid"
            val = item.split('__')
            if len(val) != 2:
                return False
            entitytypeid = val[0]
            propertyname = val[1]
            
            v = entitytypeid.split('_')
            entitytypeid = '%s.%s' % ('_'.join(v[:-1]), v[-1])
            return (entitytypeid, propertyname)

        ret = {}
        for key, value in entity.iteritems():
            r = dec(key)
            if r:
                if r[0] not in ret:
                    ret[r[0]] = {}
                ret[r[0]][r[1]] = value

        ret2 = []
        for key, value in ret.iteritems():
            value['entitytypeid'] = key
            ret2.append(value)
        return ret2



'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

from arches.management.commands import utils


class Validation_Errors(object):
    def __init__(self, *args):
        # Create lists for specific error types here and reference them when using the append_error function below.
        self.example_error_list = []


validation_errors = Validation_Errors()


# # EXAMPLE:
# def example_validation_function(resource):
# 	if resource does not pass some test:
# 		append_error('My error text', 'example_error_list')


# def append_error(text, error_type):
# 	# Use this function to append error text to an error_type_list that can be printed to a log file. 
# 	# text = the error text you would like to append. 
# 	# error_type = the error type list from the validation_errors object.
# 	error_type_list = getattr(validation_errors, error_type)
# 	error_type_list.append(text)


def validate_resource(resource):
	pass
	# Use this function to call specific validation functions.
	# example_validation_function(resource)


import inspect
import posixpath
import os
import glob
from django.template import Template
from django.conf import settings
from django.template import Context
from arches.management.commands import utils

def create_sqlfile(database_settings, path_to_file):
	db_directory = os.path.join(settings.ROOT_DIR,'db')
	context = Context(database_settings)

	def source(file):
	    return "\i \'" + file + "\'\n"

	# Generate a sql file that sources all necessary sql files into one file
	buffer = ''
	buffer = buffer + "\n-- Run all the sql scripts in the dependencies folder\n"
	for infile in glob.glob(posixpath.join(db_directory, 'install', 'dependencies', '*.sql') ):
		buffer = buffer + source(infile.replace("\\", posixpath.sep))
		    
	buffer = buffer + "\n-- Reload all managed schemas\n"
	for infile in glob.glob( posixpath.join(db_directory, 'ddl', '*.sql') ):
	    buffer = buffer + source(infile.replace("\\", posixpath.sep))

	buffer = buffer + "\n-- Add all the data in the dml folder\n"
	for infile in glob.glob( posixpath.join(db_directory, 'dml', '*.sql') ):
	    buffer = buffer + source(infile.replace("\\", posixpath.sep))
	        
	buffer = buffer + "\n-- Run all the sql in teh postdeployment folder\n"
	for infile in glob.glob( posixpath.join(db_directory, 'install', 'postdeployment', '*.sql') ):
	    buffer = buffer + source(infile.replace("\\", posixpath.sep))

	buffer = buffer + "\n-- Spring cleaning\n"
	buffer = buffer + "VACUUM ANALYZE;\n"

	utils.write_to_file(path_to_file, buffer)


import os
import codecs

def write_to_file(fileName, contents, mode='w', encoding='utf-8-sig', **kwargs):
    ensure_dir(fileName)
    file = codecs.open(fileName, mode=mode, encoding=encoding, **kwargs)
    file.write(contents)
    file.close()

def ensure_dir(f):
    d = os.path.dirname(f)
    if not os.path.exists(d):
        os.makedirs(d)



'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

from django.conf import settings
import urllib
import urllib2
import json

def find_candidates(search_string):
    query_args = { 'q': search_string, 'key': settings.BING_KEY }
    encoded_args = urllib.urlencode(query_args)
    url = 'http://dev.virtualearth.net/REST/v1/Locations?' + encoded_args
    response = json.loads(urllib2.urlopen(url).read())
    results = []
    for resource_set in response['resourceSets']:
        for feature in resource_set['resources']:
            results.append({
                'id': feature['name'],
                'text': feature['name'],
                'geometry': {
                    "type": "Point",
                    "coordinates": [
                      feature['point']['coordinates'][1],
                      feature['point']['coordinates'][0]
                    ]
                },
                'score': feature['confidence']
            })
        
    return results




'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

from django.conf import settings
from arches.app.models.resource import Resource

def livereload(request):
    return {
        'livereload_port': settings.LIVERELOAD_PORT
    }

def map_info(request):
    return {
        'map_info': {
            'x': settings.DEFAULT_MAP_X,
            'y': settings.DEFAULT_MAP_Y,
            'zoom': settings.DEFAULT_MAP_ZOOM,
            'bing_key': settings.BING_KEY,
            'map_min_zoom': settings.MAP_MIN_ZOOM,
            'map_max_zoom': settings.MAP_MAX_ZOOM
        }
    }

def resource_types(request):
    sorted_resource_types = sorted(settings.RESOURCE_TYPE_CONFIGS().items(), key=lambda v: v[1]['sort_order'])
    return {
        'resource_types': sorted_resource_types
    }

def app_name(request):
    return {
        'APP_NAME': settings.APP_NAME
    }

def user_can_edit(request):
    # need to implement proper permissions check here...
    # for now allowing all logged in users to be 'editors'
    return {
        'user_can_edit': request.user.is_authenticated()
    }




'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

class SearchEngineFactory(object):

    def create(self):
        backend = settings.SEARCH_BACKEND
        components = backend.split('.')
        classname = components[len(components)-1]
        modulename = ('.').join(components[0:len(components)-1])
        _temp = __import__(modulename, globals(), locals(), [classname], -1)
        return getattr(_temp, classname)()


'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

from arches.app.views.resources import ResourceForm
from django.utils.translation import ugettext as _

class EditHistory(ResourceForm):
    @staticmethod
    def get_info():
        return {
            'id': 'edit-history',
            'icon': 'fa-step-backward',
            'name': _('Review Edit History'),
            'class': EditHistory
        }




from setuptools import setup, find_packages  # Always prefer setuptools over distutils
from codecs import open  # To use a consistent encoding
from os import path
from setuptools.command.install import install

class post_install(install):
    def run(self):
        from arches.setup import install as arches_install         
        install.run(self)
        arches_install()

# Dynamically calculate the version based on arches.VERSION.
version = __import__('arches').__version__

setup(
    name='arches',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # http://packaging.python.org/en/latest/tutorial.html#version
    version=version,

    description='Arches is a new, open-source, web-based, geospatial information system for cultural heritage inventory and management.',
    long_description=open('README.txt').read(),
    url='http://archesproject.org/',
    author='Farallon Geographics, Inc',
    author_email='dev@fargeo.com',
    license='GNU AGPL',

    cmdclass={'install': post_install},

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],

    # What does your project relate to?
    keywords='django arches cultural heritage',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),
    include_package_data = True,
    zip_safe=False,
)


import os
import sys 
import traceback
import unicodecsv
from os import listdir
from os.path import isfile, join
from django.core import management
from django.db import connection, transaction
import concepts
from .. import utils

def load_graphs(break_on_error=True, settings=None):
    """
    Iterates through the resource node and edge files to load entitytypes and mappings into the database.
    Generates node level permissions for each resourcetype/entitytype combination

    """

    if not settings:
        from django.conf import settings        
  
    suffix = '_nodes.csv'
    errors = []
    #file_list = []

    for path in settings.RESOURCE_GRAPH_LOCATIONS:
        if os.path.exists(path):
            print '\nLOADING GRAPHS (%s)' % (path)
            print '---------------'
            for f in listdir(path):
                if isfile(join(path,f)) and f.endswith(suffix):
                    #file_list.append(join(path,f))
                    path_to_file = join(path,f)
                    basepath = path_to_file[:-10]
                    name = basepath.split(os.sep)[-1]
                    if (settings.LIMIT_ENTITY_TYPES_TO_LOAD == None or name in settings.LIMIT_ENTITY_TYPES_TO_LOAD):
                        print name
                        node_list = get_list_dict(basepath + '_nodes.csv', ['ID', 'LABEL', 'MERGENODE', 'BUSINESSTABLE'])
                        edge_list = get_list_dict(basepath + '_edges.csv', ['SOURCE', 'TARGET', 'TYPE', 'ID', 'LABEL', 'WEIGHT'])
                        mods = append_branch(os.path.join(settings.ROOT_DIR, 'management', 'resource_graphs', 'ARCHES_RECORD.E31'), node_list, edge_list)
                        node_list = mods['node_list']
                        edge_list = mods['edge_list']

                        file_errors = validate_graph(node_list, edge_list)
                        try:
                            insert_mappings(node_list, edge_list)
                            link_entitytypes_to_concepts(node_list)
                        except Exception as e:
                            file_errors.append('\nERROR: %s\n%s' % (str(e), traceback.format_exc()))
                            pass

                        if len(file_errors) > 0:
                            file_errors.insert(0, 'ERRORS IN FILE: %s\n' % (basepath))
                            file_errors.append('\n\n\n\n')
                            errors = errors + file_errors  
        else:
            errors.append('\n\nPath in settings.RESOURCE_GRAPH_LOCATIONS doesn\'t exist (%s)' % (path))                 

    utils.write_to_file(os.path.join(settings.PACKAGE_ROOT, 'logs', 'resource_graph_errors.txt'), '')
    if len(errors) > 0:
        utils.write_to_file(os.path.join(settings.PACKAGE_ROOT, 'logs', 'resource_graph_errors.txt'), '\n'.join(errors))
        print "\n\nERROR: There were errors in some of the resource graphs."
        print "Please review the errors at %s, \ncorrect the errors and then rerun this script." % (os.path.join(settings.PACKAGE_ROOT, 'logs', 'resource_graph_errors.txt'))
        if break_on_error:
            sys.exit(101)

    print '\nINDEXING ENTITY NODES'
    print '---------------------'
    concepts.index_entity_concept_lables()

    print '\nADDING NODE LEVEL PERMISSIONS'
    print '-----------------------------'
    management.call_command('packages', operation='build_permissions') 

def append_branch(path_to_branch, node_list, edge_list):
    """
    Appends a branch to the root of the passed in graph (represented as the node_list and edge_list)

    Args:
       path_to_branch (str):  path to the file where the branch graph lives(minus the _nodes.csv or _edges.csv suffix) 
            (eg: '<path>/ARCHES_RECORD.E31', where the actual nodes and edges files look like ARCHES_RECORD.E31_nodes.csv and ARCHES_RECORD.E31_edges.csv)

    Returns:
       {'node_list': node_list_with_added_branch, 'edge_list': edge_list_with_added_branch}

    """

    nodes_to_append = get_list_dict(path_to_branch + '_nodes.csv', ['ID', 'LABEL', 'MERGENODE', 'BUSINESSTABLE'])
    edges_to_append = get_list_dict(path_to_branch + '_edges.csv', ['SOURCE', 'TARGET', 'TYPE', 'ID', 'LABEL', 'WEIGHT'])

    root_node_id_of_main_graph = get_root_node_id(edge_list)
    root_node_id_of_branch = get_root_node_id(edges_to_append)
    
    # rename branch root (which is a dummy value) to the graph root value
    for edge in edges_to_append:
        if edge['SOURCE'] == root_node_id_of_branch:
            edge['SOURCE'] = root_node_id_of_main_graph

    node_list = node_list + nodes_to_append
    edge_list = edge_list + edges_to_append
    
    return {'node_list': node_list, 'edge_list': edge_list}

def get_list_dict(pathtofile, fieldnames):
    """
    Gets a list of dictionaries from a csv file

    """

    ret = []
    with open(pathtofile, 'rU') as f:
        rows = unicodecsv.DictReader(f, fieldnames=fieldnames, 
            encoding='utf-8-sig', delimiter=',', restkey='ADDITIONAL', restval='MISSING')
        rows.next() # skip header row
        for row in rows:  
            ret.append(row)
    return ret

def validate_graph(node_list, edge_list):
    """
    Validate that the graphs meet certain criteria

    """

    errors = []
    node_types = []
    nodes_referenced_in_edgefile = []
    for i in edge_list:
        nodes_referenced_in_edgefile.append(i['SOURCE'])
        nodes_referenced_in_edgefile.append(i['TARGET'])
    nodes_referenced_in_edgefile = list(set(nodes_referenced_in_edgefile))

    def append_error(text, node):
         errors.append('ERROR (%s): in node %s: %s' % (text, node['LABEL'], node))

    for node in node_list:
        # validate business table name
        if node['BUSINESSTABLE'] != u'' and node['BUSINESSTABLE'] not in [u'strings', u'dates', u'domains', u'files', u'geometries', u'numbers']:
            append_error('Incorrect business table', node)

        # validate node uniqueness
        if node['LABEL'] in node_types:
            append_error('Duplicate label', node)
            node_types.append(node['LABEL'])

        # validate no spaces in node label
        # if node['LABEL'].strip().replace(' ','') != node['LABEL']:
        #     append_error('Spaces are not allowed in labels, use underscores instead', node)

        # validate uppercase convention in node label
        if node['LABEL'].upper() != node['LABEL']:
            append_error('Use all upper case for labels', node)
    
        # validate nodes have edges
        if node['ID'] not in nodes_referenced_in_edgefile:
            append_error('Node isn\'t connected to any other node', node)

        if 'MISSING' in node:
            append_error('A row in the csv file wasn\'t parsed properly. Node: %s' % (node))

    return errors

def insert_mappings(nodes, edges):
    """
    Inserts the mappings, mapping steps, and entitytypes into the database

    """

    cursor = connection.cursor()

    for node in nodes:
        entitytypeid = get_node(node["ID"], nodes)["LABEL"]
        if node['BUSINESSTABLE'] != '' and node['BUSINESSTABLE'] != 'entities':
            mapping = get_mapping_steps_from_node_id(node["ID"], nodes, edges) 
            sql = """
                SELECT ontology.insert_mappings('%s','%s');
                """ % (", ".join(mapping), node["MERGENODE"])
            # print sql
            cursor.execute(sql)
            cursor.fetchone()

    transaction.commit_unless_managed()

def get_mapping_steps_from_node_id(node_id, nodes, edges, mapping=None):
    """
    Gets a complete mapping (from root to leaf) for a given leaf node

    """

    for edge in edges:
        if mapping == None:
            mapping = []
            node = get_node(node_id, nodes)
            mapping.append(node["LABEL"])
            mapping.append(node["BUSINESSTABLE"])
        if node_id == edge['TARGET']:
            mapping.insert(0, edge["LABEL"])
            node = get_node(edge["SOURCE"], nodes)
            mapping.insert(0, node["BUSINESSTABLE"])
            mapping.insert(0, node["LABEL"])
            get_mapping_steps_from_node_id(edge["SOURCE"], nodes, edges, mapping)
    return mapping

def get_node(id, node_list):
    for node in node_list:
        if id == node["ID"]:
            return node

def get_root_node_id(edge_list):
    target_nodes = []
    for edge in edge_list:
        target_nodes.append(edge['TARGET'])

    for edge in edge_list:
        if edge['SOURCE'] not in target_nodes:
            return edge['SOURCE']

def link_entitytypes_to_concepts(nodes):
    """
    Links entitytypes to their associated concepts

    """

    cursor = connection.cursor()
    cursor.execute("""SELECT legacyoid FROM concepts.concepts 
        WHERE conceptid = '00000000-0000-0000-0000-000000000003'
        """)
    domainlegacyid = cursor.fetchone()[0]

    cursor.execute("""SELECT legacyoid FROM concepts.concepts 
        WHERE conceptid = '00000000-0000-0000-0000-000000000004'
        """)
    otherlegacyid = cursor.fetchone()[0]

    for node in nodes:
        if node['BUSINESSTABLE'] == 'domains':
            concepts.insert_concept_relations(str(domainlegacyid), 'hasCollection', node["LABEL"])
        else:
            concepts.insert_concept_relations(str(otherlegacyid), 'hasEntity', node["LABEL"])



'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

import uuid
import copy
from operator import itemgetter
from operator import methodcaller
from django.db import transaction, connection
from django.db.models import Q
from arches.app.models import models
from arches.app.search.search_engine_factory import SearchEngineFactory
from arches.app.search.elasticsearch_dsl_builder import Match, Query
from arches.app.utils.betterJSONSerializer import JSONSerializer, JSONDeserializer

CORE_CONCEPTS = (
    '00000000-0000-0000-0000-000000000001',
    '00000000-0000-0000-0000-000000000002',
    '00000000-0000-0000-0000-000000000003'
)

class Concept(object):
    def __init__(self, *args, **kwargs):
        self.id = ''
        self.nodetype = ''
        self.legacyoid = ''
        self.relationshiptype = ''
        self.values = []
        self.subconcepts = []
        self.parentconcepts = []
        self.relatedconcepts = []
        self.hassubconcepts = False

        if len(args) != 0:
            if isinstance(args[0], basestring):
                try:
                    uuid.UUID(args[0])
                    self.get(args[0])
                except(ValueError):
                    self.load(JSONDeserializer().deserialize(args[0]))  
            elif isinstance(args[0], dict):
                self.load(args[0])  
            elif isinstance(args[0], object):
                self.load(args[0])  

    def __unicode__(self):
        return ('%s - %s') % (self.get_preflabel().value, self.id)
    
    def load(self, value):
        if isinstance(value, dict):
            self.id = value['id'] if 'id' in value else ''
            self.nodetype = value['nodetype'] if 'nodetype' in value else ''
            self.legacyoid = value['legacyoid'] if 'legacyoid' in value else ''
            self.relationshiptype = value['relationshiptype'] if 'relationshiptype' in value else ''            
            if 'values' in value:
                for val in value['values']:
                    self.addvalue(val)
            if 'subconcepts' in value:
                for subconcept in value['subconcepts']:
                    self.addsubconcept(subconcept)
            if 'parentconcepts' in value:
                for parentconcept in value['parentconcepts']:
                    self.addparent(parentconcept)
            if 'relatedconcepts' in value:
                for relatedconcept in value['relatedconcepts']:
                    self.addrelatedconcept(relatedconcept)

        if isinstance(value, models.Concepts):
            self.id = value.pk
            self.nodetype = value.nodetype_id
            self.legacyoid = value.legacyoid

    def get(self, id='', legacyoid='', include_subconcepts=False, include_parentconcepts=False, include_relatedconcepts=False, exclude=[], include=[], depth_limit=None, up_depth_limit=None, lang='en-us', **kwargs):
        if id != '':
            self.load(models.Concepts.objects.get(pk=id))
        elif legacyoid != '':
            self.load(models.Concepts.objects.get(legacyoid=legacyoid))

        if self.id != '':
            nodetype = kwargs.pop('nodetype', self.nodetype)
            uplevel = kwargs.pop('uplevel', 0)
            downlevel = kwargs.pop('downlevel', 0)
            depth_limit = depth_limit if depth_limit == None else int(depth_limit)
            up_depth_limit = up_depth_limit if up_depth_limit == None else int(up_depth_limit)

            if include != None:
                if len(include) > 0 and len(exclude) > 0:
                    raise Exception('Only include values for include or exclude, but not both')
                include = include if len(include) != 0 else models.ValueTypes.objects.distinct('category').values_list('category', flat=True)
                include = set(include).difference(exclude)
                exclude = []

                if len(include) > 0:
                    values = models.Values.objects.filter(conceptid = self.id)
                    for value in values:
                        if value.valuetype.category in include:
                            self.values.append(ConceptValue(value))

            hassubconcepts = models.ConceptRelations.objects.filter(Q(conceptidfrom = self.id), Q(relationtype__category = 'Semantic Relations'), ~Q(relationtype = 'related'))[0:1]
            if len(hassubconcepts) > 0:
                self.hassubconcepts = True
            if include_subconcepts:
                conceptrealations = models.ConceptRelations.objects.filter(Q(conceptidfrom = self.id), Q(relationtype__category = 'Semantic Relations'), ~Q(relationtype = 'related'))
                if depth_limit == None or downlevel < depth_limit:
                    if depth_limit != None:
                        downlevel = downlevel + 1                
                    for relation in conceptrealations:
                        self.relationshiptype = relation.relationtype.pk
                        self.subconcepts.append(Concept().get(id=relation.conceptidto_id, include_subconcepts=include_subconcepts, 
                            include_parentconcepts=False, include_relatedconcepts=include_relatedconcepts, exclude=exclude, include=include, depth_limit=depth_limit, 
                            up_depth_limit=up_depth_limit, downlevel=downlevel, uplevel=uplevel, nodetype=nodetype))

                    self.subconcepts = sorted(self.subconcepts, key=methodcaller('get_sortkey', lang=lang), reverse=False) 

            if include_parentconcepts:
                if nodetype == 'EntityType':
                    conceptrealations = models.ConceptRelations.objects.filter(conceptidto = self.id, relationtype = 'member')
                else:
                    conceptrealations = models.ConceptRelations.objects.filter(Q(conceptidto = self.id), Q(relationtype__category = 'Semantic Relations'), ~Q(relationtype = 'related'))
                if up_depth_limit == None or uplevel < up_depth_limit:
                    if up_depth_limit != None:
                        uplevel = uplevel + 1          
                    for relation in conceptrealations:
                        self.parentconcepts.append(Concept().get(id=relation.conceptidfrom_id, include_subconcepts=False, 
                            include_parentconcepts=include_parentconcepts, include_relatedconcepts=include_relatedconcepts, 
                            exclude=exclude, include=include, depth_limit=depth_limit, 
                            up_depth_limit=up_depth_limit, downlevel=downlevel, uplevel=uplevel, nodetype=nodetype))

            if include_relatedconcepts:
                conceptrealations = models.ConceptRelations.objects.filter(Q(relationtype = 'related') | Q(relationtype = 'member') | Q(relationtype__category = 'Mapping Properties'), Q(conceptidto = self.id) | Q(conceptidfrom = self.id))
                for relation in conceptrealations:
                    self.relationshiptype = relation.relationtype.pk
                    if relation.conceptidto_id != self.id:
                        self.relatedconcepts.append(Concept().get(relation.conceptidto_id, include=['label']))
                    if relation.conceptidfrom_id != self.id:
                        self.relatedconcepts.append(Concept().get(relation.conceptidfrom_id, include=['label']))

        return self
            
    def save(self):
        self.id = self.id if (self.id != '' and self.id != None) else str(uuid.uuid4())
        concept, created = models.Concepts.objects.get_or_create(pk=self.id, defaults={'legacyoid': self.legacyoid, 'nodetype_id': self.nodetype})

        for value in self.values:
            if not isinstance(value, ConceptValue): 
                value = ConceptValue(value)
            value.conceptid = self.id
            value.save()        

        for parentconcept in self.parentconcepts:
            parentconcept.save()
            conceptrelation = models.ConceptRelations()
            conceptrelation.pk = str(uuid.uuid4())
            conceptrelation.conceptidfrom_id = parentconcept.id 
            conceptrelation.conceptidto_id = self.id
            conceptrelation.relationtype_id = parentconcept.relationshiptype
            conceptrelation.save()

        for subconcept in self.subconcepts:
            subconcept.save()
            conceptrelation = models.ConceptRelations()
            conceptrelation.pk = str(uuid.uuid4())
            conceptrelation.conceptidfrom_id = self.id
            conceptrelation.conceptidto_id = subconcept.id
            conceptrelation.relationtype_id = subconcept.relationshiptype
            conceptrelation.save()

        for relatedconcept in self.relatedconcepts:
            relation = models.ConceptRelations()
            relation.pk = str(uuid.uuid4())
            relation.conceptidfrom_id = self.id
            relation.conceptidto_id = relatedconcept.id
            relation.relationtype_id = relatedconcept.relationshiptype
            relation.save()

        return concept

    def delete(self, delete_self=False):
        for subconcept in self.subconcepts:
            concepts_to_delete = Concept.gather_concepts_to_delete(subconcept)

            for key, concept in concepts_to_delete.iteritems():
                models.Concepts.objects.get(pk=key).delete()

        for parentconcept in self.parentconcepts:
            conceptrelations = models.ConceptRelations.objects.filter(relationtype__category = 'Semantic Relations', conceptidfrom = parentconcept.id, conceptidto = self.id)
            for relation in conceptrelations:
                relation.delete()

        deletedrelatedconcepts = []
        for relatedconcept in self.relatedconcepts:
            conceptrelations = models.ConceptRelations.objects.filter(Q(relationtype = 'related') | Q(relationtype = 'member') | Q(relationtype__category = 'Mapping Properties'), conceptidto = relatedconcept.id, conceptidfrom = self.id)
            for relation in conceptrelations:
                relation.delete()
                deletedrelatedconcepts.append(relatedconcept)

            conceptrelations = models.ConceptRelations.objects.filter(Q(relationtype = 'related') | Q(relationtype = 'member') | Q(relationtype__category = 'Mapping Properties'), conceptidfrom = relatedconcept.id, conceptidto = self.id)
            for relation in conceptrelations:
                relation.delete()
                deletedrelatedconcepts.append(relatedconcept)

        for deletedrelatedconcept in deletedrelatedconcepts:
            if deletedrelatedconcept in self.relatedconcepts:
                self.relatedconcepts.remove(deletedrelatedconcept)

        for value in self.values:
            if not isinstance(value, ConceptValue): 
                value = ConceptValue(value)
            value.delete()

        if delete_self:
            models.Concepts.objects.get(pk=self.id).delete()
        return

    @staticmethod
    def gather_concepts_to_delete(concept, lang='en-us'):
        """
        Gets a dictionary of all the concepts ids to delete
        The values of the dictionary keys differ somewhat depending on the node type being deleted
        If the nodetype == 'Concept' then return ConceptValue objects keyed to the concept id
        If the nodetype == 'ConceptScheme' or 'ConceptSchemeGroup' then return a ConceptValue object with the value set to any ONE prefLabel keyed to the concept id
        We do this because it takes so long to gather the ids of the concepts when deleting a Scheme or Group

        """

        concepts_to_delete = {}

        # Here we have to worry about making sure we don't delete nodes that have more than 1 parent
        if concept.nodetype == 'Concept':
            concept = Concept().get(id=concept.id, include_subconcepts=True, include_parentconcepts=True, include=['label'], up_depth_limit=1)
            
            def find_concepts(concept):
                if len(concept.parentconcepts) <= 1:
                    concepts_to_delete[concept.id] = concept.get_preflabel(lang=lang)
                    for subconcept in concept.subconcepts:
                        find_concepts(subconcept)

            find_concepts(concept)
            return concepts_to_delete

        # here we can just delete everything and so use a recursive CTE to get the concept ids much more quickly 
        if concept.nodetype == 'ConceptScheme' or concept.nodetype == 'ConceptSchemeGroup':
            rows = Concept().get_child_concepts(concept.id, 'narrower', ['prefLabel'], 'prefLabel')
            for row in rows:
                concepts_to_delete[row[0]] = ConceptValue({'value':row[2]})
                concepts_to_delete[row[1]] = ConceptValue({'value':row[3]})

        return concepts_to_delete

    def get_child_concepts(self, conceptid, relationtype, child_valuetypes, parent_valuetype):
        """
        Recursively builds a list of child concepts for a given concept based on its relationship type and valuetypes. 

        """

        cursor = connection.cursor()
        sql = """WITH RECURSIVE children AS (
                SELECT d.conceptidfrom, d.conceptidto, c2.value, c.value as valueto, c.valuetype, 1 AS depth       ---|NonRecursive Part
                    FROM concepts.relations d
                    JOIN concepts.values c ON(c.conceptid = d.conceptidto) 
                    JOIN concepts.values c2 ON(c2.conceptid = d.conceptidfrom) 
                    WHERE d.conceptidfrom = '{0}'
                    and c2.valuetype = '{3}'
                    and c.valuetype in ('{2}')
                    and d.relationtype = '{1}'
                UNION
                    SELECT a.conceptidfrom, a.conceptidto, v2.value, v.value as valueto, v.valuetype, depth+1      ---|RecursivePart
                    FROM concepts.relations  a
                    JOIN children b ON(b.conceptidto = a.conceptidfrom) 
                    JOIN concepts.values v ON(v.conceptid = a.conceptidto) 
                    JOIN concepts.values v2 ON(v2.conceptid = a.conceptidfrom) 
                    WHERE  v2.valuetype = '{3}'
                    and v.valuetype in ('{2}')
                    and a.relationtype = '{1}'
            ) 
            SELECT conceptidfrom, conceptidto, value, valueto FROM children;""".format(conceptid, relationtype, ("','").join(child_valuetypes), parent_valuetype)

        cursor.execute(sql)
        rows = cursor.fetchall()
        return rows

    def traverse(self, func, direction='down', scope=None):
        """
        Traverses a concept graph from self to leaf (direction='down') or root (direction='up') calling 
        the given function on each node passes an optional scope to each function

        Return a value from the function to prematurely end the traversal

        """

        if scope == None:
            ret = func(self)
        else:
            ret = func(self, scope) 
        
        # break out of the traversal if the function returns a value
        if ret != None:
            return ret    

        if direction == 'down':
            for subconcept in self.subconcepts:
                ret = subconcept.traverse(func, direction, scope)
                if ret != None: 
                    return ret  
        else:
            for parentconcept in self.parentconcepts:
                ret = parentconcept.traverse(func, direction, scope)
                if ret != None: 
                    return ret        

    def get_sortkey(self, lang='en-us'):
        for value in self.values:
            if value.type == 'sortorder':
                return value.value
                
        return self.get_preflabel(lang=lang).value

    def get_preflabel(self, lang='en-us'):
        ret = []
        if self.values == []: 
            concept = Concept().get(id=self.id, include_subconcepts=False, include_parentconcepts=False, include=['label'])
        else:
            concept = self
        for value in concept.values:
            if value.type == 'prefLabel':
                if value.language == lang:
                    return value
                elif value.language == lang.split('-')[0]:
                    ret.insert(0, value)
            elif value.type == 'altLabel':
                if value.language == lang:
                    ret.insert(0, value)
            ret.append(value)
        return ret[0] if len(ret) > 0 else ConceptValue()

    def flatten(self, ret=None):
        """
        Flattens the graph into a unordered list of concepts

        """

        if ret == None:
            ret = []

        ret.append(self)
        for subconcept in self.subconcepts:
            subconcept.flatten(ret)
            
        return ret

    def addparent(self, value):
        if isinstance(value, dict):
            self.parentconcepts.append(Concept(value))
        elif isinstance(value, Concept):
            self.parentconcepts.append(value)
        else:
            raise Exception('Invalid parent concept definition: %s' % (value))

    def addsubconcept(self, value):
        if isinstance(value, dict):
            self.subconcepts.append(Concept(value))
        elif isinstance(value, Concept):
            self.subconcepts.append(value)
        else:
            raise Exception('Invalid subconcept definition: %s' % (value))

    def addrelatedconcept(self, value):
        if isinstance(value, dict):
            self.relatedconcepts.append(Concept(value))
        elif isinstance(value, Concept):
            self.relatedconcepts.append(value)
        else:
            raise Exception('Invalid related concept definition: %s' % (value))

    def addvalue(self, value):
        if isinstance(value, dict):
            value['conceptid'] = self.id
            self.values.append(ConceptValue(value))
        elif isinstance(value, ConceptValue):
            self.values.append(value)
        elif isinstance(value, models.Values):
            self.values.append(ConceptValue(value))
        else:
            raise Exception('Invalid value definition: %s' % (value))

    def index(self, scheme=None):
        for value in self.values:
            if scheme == None:
                scheme = self.get_context()
            value.index(scheme=scheme)        

        for subconcept in self.subconcepts:
            subconcept.index(scheme=subconcept.get_context())

    def delete_index(self, delete_self=False):
        se = SearchEngineFactory().create()

        for subconcept in self.subconcepts:
            concepts_to_delete = Concept.gather_concepts_to_delete(subconcept)

            for key, concept in concepts_to_delete.iteritems():
                query = Query(se, start=0, limit=10000)
                phrase = Match(field='conceptid', query=key, type='phrase')
                query.add_query(phrase)
                query.delete(index='concept_labels')

        for value in self.values:
            value.delete_index()

    def concept_tree(self, top_concept='00000000-0000-0000-0000-000000000001', lang='en-us'):
        class concept(object):
            def __init__(self, *args, **kwargs):
                self.label = ''
                self.labelid = ''
                self.id = ''
                self.load_on_demand = False
                self.children = []    

        def _findNarrowerConcept(conceptid, depth_limit=None, level=0):
            labels = models.Values.objects.filter(conceptid = conceptid)
            ret = concept()  
            temp = Concept()        
            for label in labels:
                temp.addvalue(label)
                # if label.valuetype_id == 'prefLabel':
                #     ret.label = label.value
                #     ret.id = label.conceptid_id
                #     ret.labelid = label.valueid
            label = temp.get_preflabel(lang=lang)
            ret.label = label.value
            ret.id = label.conceptid
            ret.labelid = label.id

            conceptrealations = models.ConceptRelations.objects.filter(Q(conceptidfrom = conceptid), ~Q(relationtype = 'related'), ~Q(relationtype__category = 'Mapping Properties'))
            if depth_limit != None and len(conceptrealations) > 0 and level >= depth_limit:
                ret.load_on_demand = True
            else:
                if depth_limit != None:
                    level = level + 1                
                for relation in conceptrealations:
                    ret.children.append(_findNarrowerConcept(relation.conceptidto_id, depth_limit=depth_limit, level=level))   

            return ret 

        def _findBroaderConcept(conceptid, child_concept, depth_limit=None, level=0):
            conceptrealations = models.ConceptRelations.objects.filter(Q(conceptidto = conceptid), ~Q(relationtype = 'related'), ~Q(relationtype__category = 'Mapping Properties'))
            if len(conceptrealations) > 0 and conceptid != top_concept:
                labels = models.Values.objects.filter(conceptid = conceptrealations[0].conceptidfrom_id)
                ret = concept()          
                temp = Concept()        
                for label in labels:
                    temp.addvalue(label)
                label = temp.get_preflabel(lang=lang)
                ret.label = label.value
                ret.id = label.conceptid
                ret.labelid = label.id

                ret.children.append(child_concept)
                return _findBroaderConcept(conceptrealations[0].conceptidfrom_id, ret, depth_limit=depth_limit, level=level)
            else:
                return child_concept

        graph = []
        #if self.id == None or self.id == '' or self.id == top_concept:
        concepts = models.Concepts.objects.filter(Q(nodetype = 'ConceptSchemeGroup') | Q(nodetype = 'GroupingNode'))
        for conceptmodel in concepts:
            graph.append(_findNarrowerConcept(conceptmodel.pk, depth_limit=1))
        #else:
            #graph = [_findNarrowerConcept(self.id, depth_limit=1)]
            #concepts = _findNarrowerConcept(self.id, depth_limit=1)
            #graph = [_findBroaderConcept(self.id, concepts, depth_limit=1)]

        return graph

    def get_paths(self, lang='en-us'):
        def graph_to_paths(current_concept, path=[], path_list=[]):
            if len(path) == 0:
                current_path = []
            else:
                current_path = path[:]    
                
            current_path.insert(0, {'label': current_concept.get_preflabel(lang=lang).value, 'relationshiptype': current_concept.relationshiptype, 'id': current_concept.id})

            if len(current_concept.parentconcepts) == 0:
                path_list.append(current_path[:])
            else:    
                for parent in current_concept.parentconcepts:   
                    ret = graph_to_paths(parent, current_path, path_list)                    
            
            return path_list

        return graph_to_paths(self)

    def get_node_and_links(self, lang='en-us'):
        nodes = [{'concept_id': self.id, 'name': self.get_preflabel(lang=lang).value,'type': 'Current'}]
        links = []

        def get_parent_nodes_and_links(current_concept):
            parents = current_concept.parentconcepts
            for parent in parents:
                nodes.append({'concept_id': parent.id, 'name': parent.get_preflabel(lang=lang).value, 'type': 'Root' if len(parent.parentconcepts) == 0 else 'Ancestor'})
                links.append({'target': current_concept.id, 'source': parent.id, 'relationship': 'broader' })
                get_parent_nodes_and_links(parent)

        get_parent_nodes_and_links(self)
        for child in self.subconcepts:
            nodes.append({'concept_id': child.id, 'name': child.get_preflabel(lang=lang).value, 'type': 'Descendant' })
            links.append({'source': self.id, 'target': child.id, 'relationship': 'narrower' })

        for related in self.relatedconcepts:
            nodes.append({'concept_id': related.id, 'name': related.get_preflabel(lang=lang).value, 'type': 'Related' })
            links.append({'source': self.id, 'target': related.id, 'relationship': 'related' })

        # get unique node list and assign unique integer ids for each node (required by d3)
        nodes = {node['concept_id']:node for node in nodes}.values()
        for i in range(len(nodes)):
            nodes[i]['id'] = i
            for link in links:
                link['source'] = i if link['source'] == nodes[i]['concept_id'] else link['source']
                link['target'] = i if link['target'] == nodes[i]['concept_id'] else link['target']
        
        return {'nodes': nodes, 'links': links}

    def get_context(self):
        if self.nodetype == 'Concept' or self.nodetype == 'Collection' or self.nodetype == 'EntityType':
            concept = Concept().get(id = self.id, include_parentconcepts = True, include = None)
            print JSONSerializer().serializeToPython(concept)
            def get_scheme_id(concept):
                if concept.nodetype == 'ConceptScheme' or concept.nodetype == 'GroupingNode':
                    return concept

            return concept.traverse(get_scheme_id, direction='up')

        if self.nodetype == 'ConceptScheme' or self.nodetype == 'ConceptSchemeGroup':
            return self

    def get_e55_domain(self, entitytypeid):
        """
        For a given entitytypeid creates a dictionary for each collector node. Each collector dictionary includes
        an array of its child nodes. Returns a list of these collector dictionaries added to a list of nodes that are
        neither collectors nor children of collectors

        """

        domains = list(models.VwEntitytypeDomains.objects.filter(entitytypeid=entitytypeid).order_by('sortorder', 'value').values())
        domains_copy = copy.deepcopy(domains) #Prevents 'recursive dictionary' type from getting created in nested for loop 
        nested_domains = []
        conceptids_in_nested = []
        for domain in domains:
            if domain['valuetype'] == 'collector':
                conceptids_in_nested.append(domain['conceptid'])
                children = [] #(conceptid, relationtype, child_valuetype, parent_valuetype)
                collector_children = Concept().get_child_concepts(domain['conceptid'], 'member', ['prefLabel','collector'], 'collector')
                for child in collector_children:
                    for d in domains_copy:
                        if child[1] == d['conceptid']:
                            children.append(d)
                            conceptids_in_nested.append(d['conceptid'])
                nested_domain = domain
                nested_domain['children'] = sorted(children, key=itemgetter('sortorder'))
                nested_domain['disabled'] = True
                nested_domains.append(nested_domain)
        collectorless_domains = [a for a in domains_copy if a['conceptid'] not in conceptids_in_nested]
        result = sorted(nested_domains + collectorless_domains, key=itemgetter('sortorder')) 
        return result


class ConceptValue(object):
    def __init__(self, *args, **kwargs):
        self.id = ''
        self.conceptid = ''
        self.type = ''
        self.category = ''
        self.value = ''
        self.language = ''  

        if len(args) != 0:
            if isinstance(args[0], basestring):
                try:
                    uuid.UUID(args[0])
                    self.get(args[0])
                except(ValueError):
                    self.load(JSONDeserializer().deserialize(args[0]))  
            elif isinstance(args[0], object):
                self.load(args[0])  


    def get(self, id=''):
        self.load(models.Values.objects.get(pk = id))
        return self

    def save(self):
        self.id = self.id if (self.id != '' and self.id != None) else str(uuid.uuid4())
        value = models.Values()
        value.pk = self.id
        value.value = self.value
        value.conceptid_id = self.conceptid # models.Concepts.objects.get(pk=self.conceptid)
        value.valuetype_id = self.type # models.ValueTypes.objects.get(pk=self.type)
        if self.language != '':
            value.languageid_id = self.language # models.DLanguages.objects.get(pk=self.language)
        value.save()

    def delete(self):
        if self.id != '':
            newvalue = models.Values.objects.get(pk=self.id)
            newvalue.delete()
            self = ConceptValue()
            return self
    
    def load(self, value):
        if isinstance(value, models.Values):
            self.id = value.pk
            self.conceptid = value.conceptid.pk
            self.type = value.valuetype.pk
            self.category = value.valuetype.category
            self.value = value.value
            self.language = value.languageid_id

        if isinstance(value, dict):
            self.id = value['id'] if 'id' in value else ''
            self.conceptid = value['conceptid'] if 'conceptid' in value else ''
            self.type = value['type'] if 'type' in value else ''
            self.category = value['category'] if 'category' in value else ''
            self.value = value['value'] if 'value' in value else ''
            self.language = value['language'] if 'language' in value else ''

    def index(self, scheme=None):
        if self.category == 'label':
            se = SearchEngineFactory().create()
            data = JSONSerializer().serializeToPython(self)            
            if scheme == None:
                scheme = self.get_scheme_id()
            if scheme == None:
                raise Exception('Index of label failed.  Index type (scheme id) could not be derived from the label.')

            se.create_mapping('concept_labels', scheme.id, fieldname='conceptid', fieldtype='string', fieldindex='not_analyzed')
            se.index_data('concept_labels', scheme.id, data, 'id')
            # don't create terms for entity type concepts
            if not(scheme.id == '00000000-0000-0000-0000-000000000003' or scheme.id == '00000000-0000-0000-0000-000000000004'):
                se.index_term(self.value, self.id, scheme.id, {'conceptid': self.conceptid})
    
    def delete_index(self):      
        if self.category == 'label':
            se = SearchEngineFactory().create()
            if self.category == '':
                raise Exception('Delete index failed.  Remember to specify a category for your value. %s' % JSONSerializer().serialize(self))
            scheme = self.get_scheme_id()
            if scheme == None:
                raise Exception('Delete label index failed.  Index type (scheme id) could not be derived from the label.')
            se.delete(index='concept_labels', doc_type=scheme, id=self.id)

    def get_scheme_id(self):
        se = SearchEngineFactory().create()
        result = se.search(index='concept_labels', id=self.id)
        if result['found']:
            return Concept(result['_type'])
        else:
            return None


'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

import uuid
from django.conf import settings
from django.db import transaction, IntegrityError
from django.db.models import Q
from django.http import HttpResponse, HttpResponseNotFound, HttpResponseNotAllowed
from django.views.decorators.csrf import csrf_exempt
from django.template import RequestContext
from django.shortcuts import render_to_response
from django.contrib.auth.decorators import login_required
from arches.app.models import models
from arches.app.models.concept import Concept, ConceptValue, CORE_CONCEPTS
from arches.app.search.search_engine_factory import SearchEngineFactory
from arches.app.search.elasticsearch_dsl_builder import Bool, Match, Query, Nested, Terms, GeoShape, Range, SimpleQueryString
from arches.app.utils.betterJSONSerializer import JSONSerializer, JSONDeserializer
from arches.app.utils.JSONResponse import JSONResponse
from arches.app.utils.skos import SKOSWriter, SKOSReader

@login_required
def rdm(request, conceptid):
    lang = request.GET.get('lang', 'en-us')    
    
    languages = models.DLanguages.objects.all()

    concept_scheme_groups = []
    for concept in models.Concepts.objects.filter(nodetype = 'ConceptSchemeGroup'):
        concept_scheme_groups.append(Concept().get(id=concept.pk, include=['label']).get_preflabel(lang=lang))

    concept_schemes = []
    for concept in models.Concepts.objects.filter(nodetype = 'ConceptScheme'):
        scheme = Concept().get(id=concept.pk, include_parentconcepts=True, up_depth_limit=1, include=['label'])
        group = []
        for parent in scheme.parentconcepts:
            if parent.nodetype == 'ConceptSchemeGroup':
                group.append(parent.get_preflabel(lang=lang).value)
        concept_schemes.append({'scheme': scheme.get_preflabel(lang=lang), 'group': ','.join(group)})

    return render_to_response('rdm.htm', {
            'main_script': 'rdm',
            'active_page': 'RDM',
            'languages': languages,
            'conceptid': conceptid,
            'concept_schemes': concept_schemes,
            'concept_scheme_groups': concept_scheme_groups
        }, context_instance=RequestContext(request))

@login_required
@csrf_exempt
def concept(request, conceptid):
    f = request.GET.get('f', 'json')
    lang = request.GET.get('lang', 'en-us')
    pretty = request.GET.get('pretty', False)

    if request.method == 'GET':

        include_subconcepts = request.GET.get('include_subconcepts', 'true') == 'true'
        include_parentconcepts = request.GET.get('include_parentconcepts', 'true') == 'true'
        include_relatedconcepts = request.GET.get('include_relatedconcepts', 'true') == 'true'
        emulate_elastic_search = request.GET.get('emulate_elastic_search', 'false') == 'true'
        depth_limit = request.GET.get('depth_limit', None)
        mode = request.GET.get('mode', 'scheme')

        if f == 'html':
            depth_limit = 1
            if not conceptid:
                return render_to_response('views/rdm/concept-report.htm', {
                    'lang': lang,
                    'concept_count': models.Concepts.objects.filter(nodetype='Concept').count(),
                    'collection_count': models.Concepts.objects.filter(nodetype='Collection').count(),
                    'scheme_count': models.Concepts.objects.filter(nodetype='ConceptScheme').count(),
                    'entitytype_count': models.Concepts.objects.filter(nodetype='EntityType').count(),
                    'default_report': True
                }, context_instance=RequestContext(request))

        if f == 'skos':
            include_parentconcepts = False
            include_subconcepts = True
            depth_limit = None

        ret = []
        labels = []
        concept_graph = Concept().get(id=conceptid, include_subconcepts=include_subconcepts, 
            include_parentconcepts=include_parentconcepts, include_relatedconcepts=include_relatedconcepts,
            depth_limit=depth_limit, up_depth_limit=None, lang=lang)

        if f == 'html':
            if concept_graph.nodetype == 'Concept' or concept_graph.nodetype == 'ConceptSchemeGroup' or concept_graph.nodetype == 'ConceptScheme' or concept_graph.nodetype == 'EntityType':
                languages = models.DLanguages.objects.all()
                valuetypes = models.ValueTypes.objects.all()
                relationtypes = models.DRelationtypes.objects.all()
                prefLabel = concept_graph.get_preflabel(lang=lang).value
                for subconcept in concept_graph.subconcepts:
                    subconcept.prefLabel = subconcept.get_preflabel(lang=lang) 
                for relatedconcept in concept_graph.relatedconcepts:
                    relatedconcept.prefLabel = relatedconcept.get_preflabel(lang=lang) 
                for value in concept_graph.values:
                    if value.category == 'label':
                        labels.append(value)
                direct_parents = [parent.get_preflabel(lang=lang) for parent in concept_graph.parentconcepts]
                return render_to_response('views/rdm/concept-report.htm', {
                    'lang': lang,
                    'prefLabel': prefLabel,
                    'labels': labels,
                    'concept': concept_graph,
                    'languages': languages,
                    'valuetype_labels': valuetypes.filter(category='label'),
                    'valuetype_notes': valuetypes.filter(category='note'),
                    'valuetype_related_values': valuetypes.filter(category='undefined'),
                    'parent_relations': relationtypes.filter(category='Semantic Relations').exclude(relationtype = 'related').exclude(relationtype='broader').exclude(relationtype='broaderTransitive'),
                    'related_relations': relationtypes.filter(Q(category='Mapping Properties') | Q(relationtype = 'related')),
                    'concept_paths': concept_graph.get_paths(lang=lang),
                    'graph_json': JSONSerializer().serialize(concept_graph.get_node_and_links(lang=lang)),
                    'direct_parents': direct_parents,
                    'in_scheme_id': concept_graph.get_context()
                }, context_instance=RequestContext(request))

            else:
                languages = models.DLanguages.objects.all()
                valuetypes = models.ValueTypes.objects.all()
                relationtypes = models.DRelationtypes.objects.filter(relationtype = 'member')
                prefLabel = concept_graph.get_preflabel(lang=lang).value
                for value in concept_graph.values:
                    if value.category == 'label':
                        labels.append(value)
                direct_parents = [parent.get_preflabel(lang=lang) for parent in concept_graph.parentconcepts]
                return render_to_response('views/rdm/entitytype-report.htm', {
                    'lang': lang,
                    'prefLabel': prefLabel,
                    'labels': labels,
                    'concept': concept_graph,
                    'languages': languages,
                    'valuetype_labels': valuetypes.filter(category='label'),
                    'valuetype_notes': valuetypes.filter(category='note'),
                    'valuetype_related_values': valuetypes.filter(category='undefined'),
                    'parent_relations': relationtypes.filter(category='Semantic Relations').exclude(relationtype='related'),
                    'related_relations': relationtypes.filter(relationtype = 'member'),
                    'concept_paths': concept_graph.get_paths(lang=lang),
                    'graph_json': JSONSerializer().serialize(concept_graph.get_node_and_links(lang=lang)),
                    'direct_parents': direct_parents
                }, context_instance=RequestContext(request))

        if f == 'skos':
            skos = SKOSWriter()
            return HttpResponse(skos.write(concept_graph, format="pretty-xml"), content_type="application/xml")

        if emulate_elastic_search:
            ret.append({'_type': id, '_source': concept_graph})
        else:
            ret.append(concept_graph)       

        if emulate_elastic_search:
            ret = {'hits':{'hits':ret}} 

        return JSONResponse(ret, indent=4 if pretty else None)   

    if request.method == 'POST': 

        if len(request.FILES) > 0:
            skosfile = request.FILES.get('skosfile', None)
            imagefile = request.FILES.get('file', None)

            if imagefile:
                value = models.FileValues(valueid = str(uuid.uuid4()), value = request.FILES.get('file', None), conceptid_id = conceptid, valuetype_id = 'image', datatype = 'text', languageid_id = 'en-us')
                value.save()
                return JSONResponse(value)

            elif skosfile:
                data = JSONDeserializer().deserialize(request.POST.get('data', None))
                if data:
                    concept = Concept(data)
                    concept.save()
                    skos = SKOSReader()
                    rdf = skos.read_file(skosfile)
                    ret = skos.save_concepts_from_skos(rdf, concept_scheme_group=concept.id)
                    return JSONResponse(ret)
            
        else:
            data = JSONDeserializer().deserialize(request.body) 
            if data:
                with transaction.atomic():
                    concept = Concept(data)
                    concept.save()

                    if conceptid not in CORE_CONCEPTS:
                        concept.index()

                    return JSONResponse(concept)


    if request.method == 'DELETE':
        data = JSONDeserializer().deserialize(request.body) 

        if data:
            with transaction.atomic():
                concept = Concept(data)

                if concept.id not in CORE_CONCEPTS:
                    concept.delete_index()                
                concept.delete()

                return JSONResponse(concept)

    return HttpResponseNotFound()

@csrf_exempt
def manage_parents(request, conceptid):
    #  need to check user credentials here

    if request.method == 'POST':
        json = request.body
        if json != None:
            data = JSONDeserializer().deserialize(json)
            
            with transaction.atomic():
                if len(data['deleted']) > 0:
                    concept = Concept({'id':conceptid})
                    for deleted in data['deleted']:
                        concept.addparent(deleted)  
    
                    concept.delete()
                
                if len(data['added']) > 0:
                    concept = Concept({'id':conceptid})
                    for added in data['added']:
                        concept.addparent(added)   
            
                    concept.save()

                return JSONResponse(data)

    else:
        HttpResponseNotAllowed(['POST'])

    return HttpResponseNotFound()

@csrf_exempt
def confirm_delete(request, conceptid):
    lang = request.GET.get('lang', 'en-us') 
    concept = Concept().get(id=conceptid)
    concepts_to_delete = [concept.value for key, concept in Concept.gather_concepts_to_delete(concept, lang=lang).iteritems()]
    #return HttpResponse('<div>Showing only 50 of %s concepts</div><ul><li>%s</ul>' % (len(concepts_to_delete), '<li>'.join(concepts_to_delete[:50]) + ''))
    return HttpResponse('<ul><li>%s</ul>' % ('<li>'.join(concepts_to_delete) + ''))

@csrf_exempt
def search(request):
    se = SearchEngineFactory().create()
    searchString = request.GET['q']
    removechildren = request.GET.get('removechildren', None)
    query = Query(se, start=0, limit=100)
    phrase = Match(field='value', query=searchString.lower(), type='phrase_prefix')
    query.add_query(phrase)
    results = query.search(index='concept_labels')

    ids = []
    if removechildren != None:
        concepts = Concept().get(id=removechildren, include_subconcepts=True, include=None)
        print concepts
        def get_children(concept):
            ids.append(concept.id)

        concepts.traverse(get_children)

    newresults = []
    cached_scheme_names = {}
    for result in results['hits']['hits']:
        if result['_source']['conceptid'] not in ids:
            # first look to see if we've already retrieved the scheme name
            # else look up the scheme name with ES and cache the result
            if result['_type'] in cached_scheme_names:
                result['in_scheme_name'] = cached_scheme_names[result['_type']]
            else:
                query = Query(se, start=0, limit=100)
                phrase = Match(field='conceptid', query=result['_type'], type='phrase')
                query.add_query(phrase)
                scheme = query.search(index='concept_labels')
                for label in scheme['hits']['hits']:
                    if label['_source']['type'] == 'prefLabel':
                        cached_scheme_names[result['_type']] = label['_source']['value']
                        result['in_scheme_name'] = label['_source']['value']

            newresults.append(result)
        else:
            print result['_source']['value']

    results['hits']['hits'] = newresults
    return JSONResponse(results)

def concept_tree(request):
    lang = request.GET.get('lang', 'en-us') 
    conceptid = request.GET.get('node', None)
    concepts = Concept({'id': conceptid}).concept_tree(lang=lang)
    return JSONResponse(concepts, indent=4)

def get_preflabel_from_valueid(valueid, lang):
    se = SearchEngineFactory().create()
    concept_label = se.search(index='concept_labels', id=valueid)
    if concept_label['found']:
        return get_preflabel_from_conceptid(get_concept_label_from_valueid(valueid)['conceptid'], lang)

def get_concept_label_from_valueid(valueid):
    se = SearchEngineFactory().create()
    concept_label = se.search(index='concept_labels', id=valueid)
    if concept_label['found']:
        return concept_label['_source']

def get_preflabel_from_conceptid(conceptid, lang):
    ret = None
    se = SearchEngineFactory().create()
    query = Query(se)
    terms = Terms(field='conceptid', terms=[conceptid])
    match = Match(field='type', query='preflabel', type='phrase')
    query.add_filter(terms)
    query.add_query(match)
    preflabels = query.search(index='concept_labels')['hits']['hits'] 
    for preflabel in preflabels:
        # get the label in the preferred language, otherwise get the label in the default language
        if preflabel['_source']['language'] == lang:
            return preflabel['_source']
        if preflabel['_source']['language'] == settings.LANGUAGE_CODE:
            ret = preflabel['_source']
    return ret


import os
import inspect
import subprocess
from django.template import Template
from django.template import Context
from django.conf import settings
from arches.management.commands import utils

def create_sqlfile(database_settings, path_to_file):
	context = Context(database_settings)

	postgres_version = subprocess.check_output(["psql", "--version"])
	if int(postgres_version.split('.')[1]) >= 2:
		context['PID'] = "pid"
	else:
		context['PID'] = "procpid"

	t = Template(
	"SELECT pg_terminate_backend({{ PID }}) from pg_stat_activity where datname='{{ NAME }}';\n"
	"\n"

	"DROP DATABASE IF EXISTS {{ NAME }};\n"
	"\n"

	"CREATE DATABASE {{ NAME }}\n"
	"  WITH ENCODING='UTF8'\n"
	"       OWNER={{ USER }}\n"
	"       TEMPLATE={{POSTGIS_TEMPLATE}}\n"
	"       CONNECTION LIMIT=-1;\n"
	"\n"
	)

	utils.write_to_file(path_to_file, t.render(context));






'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

import os
import inspect
# Django settings for Arches project.

MODE = 'PROD' #options are either "PROD" or "DEV" (installing with Dev mode set, get's you extra dependencies)
DEBUG = True
TEMPLATE_DEBUG = DEBUG
INTERNAL_IPS = ('127.0.0.1',)

#########################################
###  START PACKAGE SPECIFIC SETTINGS  ###
#########################################

DATABASES = {
    'default': {
        'ENGINE': 'django.contrib.gis.db.backends.postgis', # Add 'postgresql_psycopg2', 'postgresql', 'mysql', 'sqlite3' or 'oracle'.
        'NAME': 'arches',                      # Or path to database file if using sqlite3.
        'USER': 'postgres',                      # Not used with sqlite3.
        'PASSWORD': 'postgis',                  # Not used with sqlite3.
        'HOST': 'localhost',                      # Set to empty string for localhost. Not used with sqlite3.
        'PORT': '5432',                      # Set to empty string for default. Not used with sqlite3.
        'SCHEMAS': 'public,data,ontology,concepts', # syncdb will put the admin tables in the first listed schema,
        'POSTGIS_TEMPLATE': 'template_postgis_20',
    }
}

ENTITY_MODEL = {
    # override this setting in your packages settings.py file
    # to set the default model for the system to use
    # Your model needs to inherit from 'arches.app.models.entity.Entity' to work
    'default': 'arches.app.models.entity.Entity'
}


ELASTICSEARCH_HTTP_PORT = 9200 # this should be in increments of 200, eg: 9400, 9600, 9800
SEARCH_BACKEND = 'arches.app.search.search.SearchEngine'
# see http://elasticsearch-py.readthedocs.org/en/master/api.html#elasticsearch.Elasticsearch
ELASTICSEARCH_HOSTS = [
    {'host': 'localhost', 'port': ELASTICSEARCH_HTTP_PORT}
]
ELASTICSEARCH_CONNECTION_OPTIONS = {'timeout': 30}


SEARCH_ITEMS_PER_PAGE = 5
SEARCH_DROPDOWN_LENGTH = 100
WORDS_PER_SEARCH_TERM = 10 # set to None for unlimited number of words allowed for search terms

DISPLAY_NAME_FOR_UNNAMED_ENTITIES = 'Unnamed Resource' # override this setting in your packages settings.py file

# override this setting in your packages settings.py file
# entity type that holds the spatial coordinates of resources
ENTITY_TYPE_FOR_MAP_DISPLAY = ''

LIMIT_ENTITY_TYPES_TO_LOAD = None #(
    # override this setting in your packages settings.py file
#    'ARCHAEOLOGICAL HERITAGE (ARTIFACT).E18',
#)

DATA_CONCEPT_SCHEME = ''

ETL_USERNAME = 'ETL' # override this setting in your packages settings.py file

LIVERELOAD_PORT = 35729 # usually only used in development, 35729 is default for livereload browser extensions

DEFAULT_MAP_X = 0
DEFAULT_MAP_Y = 0
DEFAULT_MAP_ZOOM = 0
MAP_MIN_ZOOM = 0.0005831682455839253
MAP_MAX_ZOOM = 156543.03392804097
MAP_LAYER_FEATURE_LIMIT = 100000

BING_KEY = 'Ak-dzM4wZjSqTlzveKz5u0d4IQ4bRzVI309GxmkgSVr1ewS6iPSrOvOKhA-CJlm3'

RESOURCE_TYPE_CONFIGS = {
    # override this setting in your packages settings.py file
    #
    # 'HERITAGE_RESOURCE.E18': {
    #     'resourcetypeid': 'HERITAGE_RESOURCE.E18',
    #     'name': _('Heritage Resource'),
    #     'icon_class': 'fa fa-trophy',
    #     'default_page': 'summary',
    #     'description': _('INSERT RESOURCE DESCRIPTION HERE'),
    #     'categories': [_('Resource')],
    #     'has_layer': True,
    #     'on_map': True,
    #     'vector_color': '#3366FF',
    #     'primary_name_lookups': {
    #         'entity_type': 'NAME.E41',
    #         'lookup_value': 'Primary'
    #     },
    #     'sort_order': 1
    # },
}

GEOCODING_PROVIDER = 'arches.app.utils.bing_geocoder'

EXPORT_CONFIG = ''

APP_NAME = 'Arches v3.0'

#######################################
###  END PACKAGE SPECIFIC SETTINGS  ###
#######################################

ROOT_DIR = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))

ADMINS = (
    # ('Your Name', 'your_email@example.com'),
)
MANAGERS = ADMINS

POSTGIS_VERSION = (2, 0, 0)

SITE_ID = 1

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# On Unix systems, a value of None will cause Django to use the same
# timezone as the operating system.
# If running in a Windows environment this must be set to the same as your
# system time zone.
TIME_ZONE = 'America/Chicago'
USE_TZ = False

# Default Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html

# see https://docs.djangoproject.com/en/1.6/topics/i18n/translation/#how-django-discovers-language-preference 
# to see how LocaleMiddleware tries to determine the user's language preference 
# (make sure to check your accept headers as they will override the LANGUAGE_CODE setting!)
# also see get_language_from_request in django.utils.translation.trans_real.py
# to see how the language code is derived in the actual code
LANGUAGE_CODE = 'en-us'

# the path where your translation strings are stored
LOCALE_PATHS = (
    os.path.join(ROOT_DIR, 'locale'),
)

# If you set this to False, Django will not format dates, numbers and
# calendars according to the current locale
USE_L10N = True

# Absolute filesystem path to the directory that will hold user-uploaded files.
MEDIA_ROOT =  os.path.join(ROOT_DIR, 'uploadedfiles')

# URL that handles the media served from MEDIA_ROOT, used for managing stored files. 
# It must end in a slash if set to a non-empty value.
MEDIA_URL = 'http://localhost:8000/media/'

# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/home/media/media.lawrence.com/static/"
STATIC_ROOT = ''

# URL prefix for static files.
# Example: "http://media.lawrence.com/static/"
STATIC_URL = '/media/'

# URL prefix for admin static files -- CSS, JavaScript and images.
# Make sure to use a trailing slash.
# Examples: "http://foo.com/static/admin/", "/static/admin/".
ADMIN_MEDIA_PREFIX = '/media/admin/'

# Additional locations of static files
STATICFILES_DIRS = (
    # Put strings here, like "/home/html/static" or "C:/www/django/static".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
    os.path.join(ROOT_DIR, 'app', 'media'),
    MEDIA_ROOT
)

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
#    'django.contrib.staticfiles.finders.DefaultStorageFinder',
)

# Make this unique, and don't share it with anybody.
SECRET_KEY = 'c7ky-mc6vdnv+avp0r@(a)8y^51ex=25nogq@+q5$fnc*mxwdi'

# List of callables that know how to import templates from various sources.
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
#     'django.template.loaders.eggs.Loader',
)


# Application definition

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.gis',
    'arches',
    'arches.app.models',
    'arches.management'
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    #'django.middleware.locale.LocaleMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'arches.urls'

WSGI_APPLICATION = 'arches.wsgi.application'

TEMPLATE_DIRS = (
    # Put strings here, like "/home/html/django_templates" or "C:/www/django/templates".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
    os.path.join(ROOT_DIR, 'app', 'templates'),

    # Adding a reference to error page templates because of issues we were seeing in amazon aws instances
    # http://stackoverflow.com/questions/13284443/django-template-error-500-html-on-amazon-ec2?rq=1
    # os.path.join(ROOT_DIR, 'virtualenv/ENV/Lib/site-packages/django/contrib/admin/templates/admin'),
    # os.path.join(ROOT_DIR, 'virtualenv/ENV/lib/python2.7/site-packages/django/contrib/admin/templates/admin'),
)

# List of processors used by RequestContext to populate the context.
# Each one should be a callable that takes the request object as its
# only parameter and returns a dictionary to add to the context.
TEMPLATE_CONTEXT_PROCESSORS = (
    'django.contrib.auth.context_processors.auth',
    'django.core.context_processors.debug',
    'django.core.context_processors.i18n',
    'django.core.context_processors.media',
    'django.core.context_processors.static',
    'django.core.context_processors.tz',
    'django.core.context_processors.request',
    'django.contrib.messages.context_processors.messages',
    'arches.app.utils.context_processors.livereload',
    'arches.app.utils.context_processors.resource_types',
    'arches.app.utils.context_processors.map_info',
    'arches.app.utils.context_processors.app_name',
    'arches.app.utils.context_processors.user_can_edit',
)


LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'file': {
            'level': 'DEBUG',
            'class': 'logging.FileHandler',
            'filename': os.path.join(ROOT_DIR, 'arches.log'),
        },
    },
    'loggers': {
        'arches': {
            'handlers': ['file'],
            'level': 'DEBUG',
            'propagate': True,
        },
    },
}

LOGIN_URL = 'auth'

# Package specific validation.
# Should be over-written in the package settings file.
PACKAGE_VALIDATOR = 'arches.app.utils.mock_package_validator'

# Bounding box for geometry data validation. By default set to coordinate system bounding box.
# NOTE: This is not used by the front end of the application.
DATA_VALIDATION_BBOX = [(-180,-90), (-180,90), (180,90), (180,-90), (-180,-90)]

try:
    from settings_local import *
except ImportError:
    pass



import os
import csv
import datetime
from time import time
from django.conf import settings
from django.db import connection, transaction
from django.contrib.auth.models import User
from django.forms.models import model_to_dict
from django.core.management.base import BaseCommand, CommandError
from arches.app.models.entity import Entity
from arches.app.models.resource import Resource
from arches.app.models.models import Concepts
from arches.app.models.models import Values
from arches.app.models.models import RelatedResource
from arches.app.models.concept import Concept
from arches.app.search.search_engine_factory import SearchEngineFactory
from arches.management.commands import utils
from optparse import make_option
from formats.archesfile import ArchesReader
from formats.shapefile import ShapeReader


class ResourceLoader(object):

    def __init__(self):
        self.user = User()
        self.user.first_name = settings.ETL_USERNAME
        self.resources = []
        self.se = SearchEngineFactory().create()

    option_list = BaseCommand.option_list + (
        make_option('--source',
            action='store',
            dest='source',
            default='',
            help='.arches file containing resource records'),
         make_option('--format',
            action='store_true',
            default='arches',
            help='format extension that you would like to load: arches or shp'),
        )

    def load(self, source):
        file_name, file_format = os.path.splitext(source)
        if file_format == '.shp':
            reader = ShapeReader()
        elif file_format == '.arches':
            reader = ArchesReader()
            print '\nVALIDATING ARCHES FILE ({0})'.format(source)
            print '-----------------------'
            reader.validate_file(source)

        start = time()
        resources = reader.load_file(source)

        print '\nLOADING RESOURCES ({0})'.format(source)
        print '-----------------------'
        relationships = None
        related_resource_records = []
        relationships_file = file_name + '.relations'
        elapsed = (time() - start)
        print 'time to parse {0} resources = {1}'.format(file_name, elapsed)
        results = self.resource_list_to_entities(resources)
        if os.path.exists(relationships_file):
            relationships = csv.DictReader(open(relationships_file, 'r'), delimiter='|')
            for relationship in relationships:
                related_resource_records.append(self.relate_resources(relationship, results['legacyid_to_entityid']))
        else:
            print 'No relationship file'

        #self.se.bulk_index(self.resources)


    def resource_list_to_entities(self, resource_list):
        '''Takes a collection of imported resource records and saves them as arches entities'''

        start = time()
        d = datetime.datetime.now()
        load_id = 'LOADID:{0}-{1}-{2}-{3}-{4}-{5}'.format(d.year, d.month, d.day, d.hour, d.minute, d.microsecond) #Should we append the timestamp to the exported filename?

        ret = {'successfully_saved':0, 'failed_to_save':[]}
        schema = None
        current_entitiy_type = None
        legacyid_to_entityid = {}
        errors = []
        progress_interval = 250
        for count, resource in enumerate(resource_list):

            if count >= progress_interval and count % progress_interval == 0:
                print count, 'of', len(resource_list), 'loaded'

            masterGraph = None
            entityData = []
            if current_entitiy_type != resource.entitytypeid:
                schema = Resource.get_mapping_schema(resource.entitytypeid)

            master_graph = self.build_master_graph(resource, schema)

            self.pre_save(master_graph)
            master_graph.save(user=self.user, note=load_id)
            master_graph.index()
            resource.entityid = master_graph.entityid
            legacyid_to_entityid[resource.resource_id] = master_graph.entityid
            
            ret['successfully_saved'] += 1
            # self.resources.append({
            #     '_index': 'entity',
            #     '_type': master_graph.entitytypeid,
            #     '_id': master_graph.entityid,
            #     '_source': master_graph.prepare_documents_for_search_index()[0]
            # })
            

        ret['legacyid_to_entityid'] = legacyid_to_entityid
        elapsed = (time() - start)
        print len(resource_list), 'resources loaded'
        if len(resource_list) > 0:
            print 'total time to etl = %s' % (elapsed)
            print 'average time per entity = %s' % (elapsed/len(resource_list))
            print 'Load Identifier =', load_id
            print '***You can reverse this load with the following command:'
            print 'python hip remove_resources --load_id', load_id
        return ret

    def build_master_graph(self, resource, schema):
        master_graph = None
        entity_data = []
        for row in resource.nongroups:
            entity = Resource()
            entity.create_from_mapping(row.resourcetype, schema[row.attributename]['steps'], row.attributename, row.attributevalue)
            entity_data.append(entity)

        if len(entity_data) > 0:
            master_graph = entity_data[0]
            for mapping in entity_data[1:]:
                master_graph.merge(mapping)

        for group in resource.groups:
            entity_data2 = []
            for row in group.rows:
                entity = Resource()
                entity.create_from_mapping(row.resourcetype, schema[row.attributename]['steps'], row.attributename, row.attributevalue)
                entity_data2.append(entity)  

            mapping_graph = entity_data2[0]
            for mapping in entity_data2[1:]:
                mapping_graph.merge(mapping)

            if master_graph == None:
                master_graph = mapping_graph
            else:
                node_type_to_merge_at = schema[row.attributename]['mergenodeid']
                master_graph.merge_at(mapping_graph, node_type_to_merge_at)

        return master_graph

    def pre_save(self, master_graph):
        pass

    def relate_resources(self, relationship, legacyid_to_entityid):
        relationshiptype_concept = Concepts.objects.get(legacyoid = relationship['RELATION_TYPE'])
        concept_value = Values.objects.filter(conceptid = relationshiptype_concept.conceptid).filter(valuetype = 'prefLabel')
        start_date = None if relationship['START_DATE'] == '' else relationship['START_DATE']
        end_date = None if relationship['END_DATE'] == '' else relationship['END_DATE']
        related_resource_record = RelatedResource(
            entityid1 = legacyid_to_entityid[relationship['RESOURCEID_FROM']],
            entityid2 = legacyid_to_entityid[relationship['RESOURCEID_TO']],
            notes = relationship['NOTES'],
            relationshiptype = concept_value[0].valueid,
            datestarted = start_date,
            dateended = end_date,
            )
        related_resource_record.save()
        self.se.index_data(index='resource_relations', doc_type='all', body=model_to_dict(related_resource_record), idfield='resourcexid')


from django.conf import settings
from django.db import connection, transaction
from arches.app.models.resource import Resource
import arches.app.models.models as archesmodels
from django.db.models import Q
from django.db.models import Count
from optparse import make_option
from django.core.management.base import BaseCommand, CommandError
from django.core.exceptions import ObjectDoesNotExist


def delete_resources(load_id):
    """Takes the load id stored in the note column of the edit log and deletes each resource with that id"""
    resources_for_removal = archesmodels.EditLog.objects.filter( Q(note=load_id) )
    resourceids = set([editlog.resourceid for editlog in resources_for_removal])
    for r_id in resourceids:
        try:
            resource = Resource(r_id)
            resource.delete_index()
            note = '{0} Deleted'.format(load_id)
            resource.delete_all_resource_relationships()
            resource.delete(note=note)
        except ObjectDoesNotExist:
            print 'Entity does not exist. Nothing to delete'


def truncate_resources():
    """Deletes ALL resources in your database. Use with caution!"""
    cursor = connection.cursor()
    cursor.execute("""TRUNCATE data.entities CASCADE;""" )
    print 'Resources Truncated'


from django.contrib.gis.geos import MultiPoint
from django.contrib.gis.geos import MultiLineString
from django.contrib.gis.geos import MultiPolygon

def convert_geom(geos_geom):
  if geos_geom.geom_type == 'Point':
      multi_geom = MultiPoint(geos_geom)
      shp_geom = [[c for c in multi_geom.coords]]
  if geos_geom.geom_type == 'LineString':
      multi_geom = MultiLineString(geos_geom)
      shp_geom = [c for c in multi_geom.coords]
  if geos_geom.geom_type == 'Polygon':
      multi_geom = MultiPolygon(geos_geom)
      shp_geom = [c[0] for c in multi_geom.coords]
  if geos_geom.geom_type == 'MultiPoint':
      shp_geom = [[c for c in geos_geom.coords]]
  if geos_geom.geom_type == 'MultiLineString':
      shp_geom = [c for c in geos_geom.coords]
  if geos_geom.geom_type == 'MultiPolygon':
      shp_geom = [c[0] for c in geos_geom.coords]

  return shp_geom


import subprocess
from django.conf import settings

def execute_sql_file(pathtofile, database='default'):
	database = settings.DATABASES[database]
	proc = subprocess.Popen(["psql", "-h", database['HOST'], "-p", database['PORT'], "-U", database['USER'], "-d", database['NAME'], "-f", pathtofile],stdin=subprocess.PIPE,stdout=subprocess.PIPE)
	output, errors = proc.communicate()
	return output